#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri May 10 10:55:51 2019

@author: mishugeb
"""
import numpy as np
from numpy import linalg as LA
from scipy.optimize import nnls
from scipy.optimize import minimize
from handygenome.sigprofiler_clone.SigProfilerAssignment import decompose_subroutines as sub
import scipy.stats
import copy 
import os

#############################################################################################################
#################################### Functions For Single Sample Algorithms #############################

################################################################### FUNCTION ONE ###################################################################
#function to calculate the cosine similarity
def cos_sim(a, b):
      
    
    """Takes 2 vectors a, b and returns the cosine similarity according 
    to the definition of the dot product
    
    Dependencies: 
    *Requires numpy library. 
    *Does not require any custom function (constructed by me)
    
    Required by:
    * pairwise_cluster_raw
    	"""
    if np.sum(a)==0 or np.sum(b) == 0:
        return 0.0      
    dot_product = np.dot(a, b)
    norm_a = np.linalg.norm(a)
    norm_b = np.linalg.norm(b)
    return dot_product / (norm_a * norm_b)

def add_connected_sigs(background_sigs, allsigids): 
    connected_sigs = [["SBS2", "SBS13"],
                  ["SBS7a", "SBS7b", "SBS7c", "SBS7d"],
                  ["SBS10a", "SBS10b"],
                  ["SBS17a", "SBS17b"]]
    
    backround_sig_names = sub.get_items_from_index(allsigids,background_sigs )
    connect_the_sigs = []
    for i in range(len(connected_sigs)):
        if len(set(connected_sigs[i]).intersection(set(backround_sig_names)))>0:
            connect_the_sigs = connect_the_sigs+connected_sigs[i]
    
    backround_sig_names = list(set(backround_sig_names).union(set(connect_the_sigs)))    
    background_sigs = sub.get_indeces(allsigids, backround_sig_names)
    background_sigs.sort()
    return background_sigs

def parameterized_objective2_custom(x, signatures, samples):
    rec = np.dot(signatures, x)
    try:
        y = LA.norm(samples-rec[:,np.newaxis])
    except:
        y = LA.norm(samples-rec)
    return y


def constraints1(x, samples):
    sumOfSamples=np.sum(samples, axis=0)
    #print(sumOfSamples)
    #print(x)
    result = sumOfSamples-(np.sum(x))
    #print (result)
    return result


def create_bounds(idxOfZeros, samples, numOfSignatures):
    total = np.sum(samples)
    b = (0.0, float(total))
    lst =[b]*numOfSignatures
    
    for i in idxOfZeros:
        lst[i] = (0.0, 0.0) 
    
    return lst 

# required for removing signatures with background
def get_changed_background_sig_idx(exposures, background_sigs):
    
    background_sigs_values = sub.get_items_from_index(exposures,background_sigs)
    temp_exposures = exposures[:]
    temp_exposures[:] = (value for value in temp_exposures if value != 0)
    
    # remove the background signatures with zero values
    background_sigs[:] = (value for value in background_sigs_values if value != 0)
    
    # get the new indices of the background signatures 
    background_sigs = sub.get_indeces(temp_exposures, background_sigs_values)
    
    return background_sigs


# Fit signatures 
def fit_signatures(W, genome, metric="l2"):
    genome = np.array(genome)
    maxmutation = round(np.sum(genome))
     
    
    
    #initialize the guess
    x0 = np.random.rand(W.shape[1], 1)*maxmutation
    x0= x0/np.sum(x0)*maxmutation
    
    #the optimization step
    #set the bounds and constraints
    #bnds = create_bounds([], genome, W.shape[1]) 
    #cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]}
    
    
    #sol = scipy.optimize.minimize(parameterized_objective2_custom, x0, args=(W, genome),  bounds=bnds, constraints =cons1, tol=1e-30)
    #solution = sol.x
    
    
    ### using NNLS algorithm 
    reg = nnls(W,genome)
    weights = reg[0]
    est_genome = np.dot(W, weights)
    normalised_weights = weights/sum(weights)
    solution = normalised_weights*sum(genome)
    
    
    
    #print(W1)
    #convert the newExposure vector into list type structure
    newExposure = list(solution)
    
    
    # get the maximum value of the new Exposure
    maxcoef = max(newExposure)
    idxmaxcoef = newExposure.index(maxcoef)
    
    newExposure = np.round(newExposure)
    
    # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
    if np.sum(newExposure)!=maxmutation:
        newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
     
    
    
    if metric=="cosine":
        newSimilarity = cos_sim(genome, est_genome)
    else:
        newSimilarity = np.linalg.norm(genome-est_genome , ord=2)/np.linalg.norm(genome, ord=2)
    
    return (newExposure, newSimilarity)

# to do the calculation, we need: W, samples(one column), H for the specific sample, tolerence
# Fit signatures with multiprocessing
def fit_signatures_pool(total_genome, W, index, metric="l2"):
    genome = total_genome[:,index]
    genome = np.array(genome)
    maxmutation = round(np.sum(genome))
     
    
    
    #initialize the guess
    x0 = np.random.rand(W.shape[1], 1)*maxmutation
    x0= x0/np.sum(x0)*maxmutation
    
    #the optimization step
    #set the bounds and constraints
    #bnds = create_bounds([], genome, W.shape[1]) 
    #cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]}
    
    
    #sol = scipy.optimize.minimize(parameterized_objective2_custom, x0, args=(W, genome),  bounds=bnds, constraints =cons1, tol=1e-30)
    #solution = sol.x
    
    
    ### using NNLS algorithm 
    reg = nnls(W,genome)
    weights = reg[0]
    est_genome = np.dot(W, weights)
    normalised_weights = weights/sum(weights)
    solution = normalised_weights*sum(genome)
    
    #print(W1)
    #convert the newExposure vector into list type structure
    newExposure = list(solution)
    
    
    # get the maximum value of the new Exposure
    maxcoef = max(newExposure)
    idxmaxcoef = newExposure.index(maxcoef)
    
    newExposure = np.round(newExposure)
    
    # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
    if np.sum(newExposure)!=maxmutation:
        newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
     
    
    
    if metric=="cosine":
        newSimilarity = cos_sim(genome, est_genome)
    else:
        newSimilarity = np.linalg.norm(genome-est_genome , ord=2)/np.linalg.norm(genome, ord=2)
    
    return (newExposure, newSimilarity)


def add_signatures(W, genome, cutoff=0.05, presentSignatures=[], toBeAdded="all", metric="l2", solver="nnls", check_rule_negatives=[], check_rule_penalty=1.0, verbose = False):
     # This function takes an array of signature and a single genome as input, returns a dictionray of cosine similarity, exposures and presence 
     # of signatures according to the indices of the original signature array
    
    if toBeAdded == "all":
        notToBeAdded=[]
    else:               
        #print(len(list(range(W.shape[1]))))
        #print(len(toBeAdded))
        notToBeAdded = list(set(list(range(W.shape[1])))-set(toBeAdded)) # get the indices of the signatures to be excluded
        
        #print(len(notToBeAdded))
    originalSimilarity = np.inf # it can be also written as oldsimilarity. wanted to put a big number
    maxmutation = round(np.sum(genome))
    init_listed_idx = presentSignatures 
    
    init_nonlisted_idx = list(set(list(range(W.shape[1])))-set(init_listed_idx)-set(notToBeAdded))
    
    finalRecord = [["similarity place-holder" ], ["newExposure place-holder"], ["signatures place-holder"]] #for recording the cosine difference, similarity, the new exposure and the index of the best signauture
    
    # get the initial original similarity if some signatures are already given to be present
    if len(init_listed_idx)!=0:
        
        loop_liststed_idx=init_listed_idx
        loop_liststed_idx.sort()
        
        W1 = W[:,loop_liststed_idx]
       
        #print (W1.shape)
        #initialize the guess
        x0 = np.random.rand(W1.shape[1], 1)*maxmutation
        x0= x0/np.sum(x0)*maxmutation
        
        #set the bounds and constraints
        bnds = create_bounds([], genome, W1.shape[1]) 
        cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]} 
        
        #the optimization step
        if solver == "slsqp":
            sol = minimize(parameterized_objective2_custom, x0, args=(W1, genome),  bounds=bnds, constraints =cons1, tol=1e-30)
            newExposure = list(sol.x)
            est_genome = np.dot(W1, np.array(newExposure))
        if solver == "nnls":
        ### using NNLS algorithm        
            reg = nnls(W1, genome[:,0])
            weights = reg[0]
            est_genome = np.dot(W1, weights)
            normalised_weights = weights/sum(weights)
            solution = normalised_weights*sum(np.sum(genome, axis=0))
            newExposure = list(solution)
        #print(W1)
        #convert the newExposure vector into list type structure
        
        #newExposure = list(solution) #for nnls
                           
                          
        # get the maximum value of the new Exposure
        maxcoef = max(newExposure)
        idxmaxcoef = newExposure.index(maxcoef)
        
        newExposure = np.round(newExposure)
        
        # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
        if np.sum(newExposure)!=maxmutation:
            newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
         
        
        if metric=="cosine":
            originalSimilarity = 1-cos_sim(genome[:,0], est_genome )
        elif metric=="l2":    
            originalSimilarity = np.linalg.norm(genome[:,0]-est_genome , ord=2)/np.linalg.norm(genome[:,0], ord=2)
        
            
        finalRecord = [originalSimilarity, newExposure, init_listed_idx]
        
        
    while True:
        
        bestDifference = -np.inf  #### wanted to put a big initila number
        bestSimilarity = np.inf #### actually bestdistance
        loopRecord = [["newExposure place-holder"], ["signatures place-holder"], ["best loop signature place-holder"]]
        
        for sig in init_nonlisted_idx:
            
            
            
            if len(init_listed_idx)!=0:
                loop_liststed_idx=init_listed_idx+[sig]
                loop_liststed_idx.sort()
                #print(loop_liststed_idx)
                W1 = W[:,loop_liststed_idx]
                #print (W1.shape)
                #initialize the guess
                x0 = np.random.rand(W1.shape[1], 1)*maxmutation
                x0= x0/np.sum(x0)*maxmutation
                
                #set the bounds and constraints
                bnds = create_bounds([], genome, W1.shape[1]) 
                cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]} 
            # for the first time addintion  
            else:
                W1 = W[:,sig][:,np.newaxis]
                #print (W1.shape)        
                #initialize the guess
                x0 = np.ones((1,1))*maxmutation    
            
                #set the bounds and constraints
                bnds = create_bounds([], genome, 1) 
                cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genome]} 
            
            if solver == "slsqp":
                #the optimization step
                sol = minimize(parameterized_objective2_custom, x0, args=(W1, genome),  bounds=bnds, constraints =cons1, tol=1e-30)
                newExposure = list(sol.x)
                # compute the estimated genome
                est_genome = np.dot(W1, np.array(newExposure))
            if solver == "nnls":
                reg = nnls(W1, genome[:,0])
                weights = reg[0]
                est_genome = np.dot(W1, weights)
                normalised_weights = weights/sum(weights)
                solution = normalised_weights*sum(genome)
                newExposure = list(solution) #for nnls
            #print(W1)
            #convert the newExposure vector into list type structure
            
            
            
            # get the maximum value of the new Exposure
            maxcoef = max(newExposure)
            idxmaxcoef = newExposure.index(maxcoef)
            
            newExposure = np.round(newExposure)
            
            # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
            if np.sum(newExposure)!=maxmutation:
                newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
             
            
            
            if metric=='cosine':
                newSimilarity = 1-cos_sim(genome[:,0], est_genome)
            elif metric=='l2':
                newSimilarity = np.linalg.norm(genome[:,0]-est_genome , ord=2)/np.linalg.norm(genome[:,0], ord=2)
            
            if sig in check_rule_negatives:
                
                newSimilarity = newSimilarity*check_rule_penalty
                
            difference = originalSimilarity -  newSimilarity 
            
            # record the best values so far that creates the best difference
            if difference>bestDifference:
                bestDifference = difference
                bestSimilarity = newSimilarity
                loopRecord = [newExposure, W1, sig]  #recording the cosine difference, the new exposure and the index of the best signauture
                
        
         
         
        if originalSimilarity - bestSimilarity>cutoff: 
            
            
            originalSimilarity = bestSimilarity
            init_listed_idx.append(loopRecord[2])  #add the qualified signature in the signature list
            init_nonlisted_idx.remove(loopRecord[2]) #remover the qualified signature from the to be added list
            init_listed_idx.sort()
            #print(originalSimilarity)
            finalRecord = [originalSimilarity, loopRecord[0], init_listed_idx, loopRecord[1], genome]
            if verbose==True:
                print("Current signaure indices: ", init_listed_idx)
                if metric == "cosine":
                    print("Cosine similarity: ", 1-originalSimilarity)
                elif metric == "l2":
                    print("L2 norm: ", originalSimilarity)
                print("\n")
            if len(init_nonlisted_idx)!= 0:
                
                continue
            else:
                break
        else:
            
            break
     
    #print("Good so far")
    #print(finalRecord)
    
    dictExposure= {"similarity":finalRecord[0], "exposures":finalRecord[1], "signatures": finalRecord[2]}  
    addExposure = np.zeros([W.shape[1]])
    addExposure[dictExposure["signatures"]]=dictExposure["exposures"]
    #print(cos_sim(genome[:,0], np.dot(W,addExposure)))
    if metric == "cosine":
        return  addExposure, cos_sim(genome[:,0], np.dot(W,addExposure)), cos_sim(genome[:,0], np.dot(W,addExposure)) #first value is the exprosure, second value is the similarity (based on the similarity matic), third value is the cosine similarity
    elif metric == "l2":
        return  addExposure, originalSimilarity, cos_sim(genome[:,0], np.dot(W,addExposure)) #first value is the exprosure, second value is the similarity (based on the similarity matic), third value is the cosine similarity







def remove_all_single_signatures(W, H, genomes, metric="l2", solver = "nnls", cutoff=0.05, background_sigs = [], verbose=False):
    # make the empty list of the successfull combinations
    signature_ids=sub.make_letter_ids(idlenth = W.shape[1], mtype = "Signature ")
    
    current_signatures=signature_ids
    #print(current_signatures)
    successList = [0,[],0] 
    background_sig = copy.deepcopy(background_sigs)
    
    #setting the base_similarity
    base_H_index = list(np.nonzero(H)[0])
    reg = nnls(W[:,base_H_index],genomes)
    base_weights = reg[0]
    
    if metric == "cosine":
        # get the cos_similarity with sample for the oringinal W and H[:,i]
        originalSimilarity= 1-cos_sim(genomes, np.dot(W[:,base_H_index], base_weights))
        if verbose==True:
            print("originalSimilarity", 1-originalSimilarity)
    elif metric == "l2":
       
        originalSimilarity = np.linalg.norm(genomes-np.dot(W[:,base_H_index], base_weights) , ord=2)/np.linalg.norm(genomes, ord=2)
        if verbose==True:
            print("originalSimilarity", originalSimilarity)
    # make the original exposures of specific sample round
    oldExposures = np.round(H)
    
    # set the flag for the while loop
    if len(oldExposures[np.nonzero(oldExposures)])>1:
        Flag = True
    else: 
        Flag = False
        if metric=="cosine":
            return oldExposures, 1-originalSimilarity, cos_sim(genomes, np.dot(W,H))  #first value is the exprosure, second value is the similarity (based on the similarity matic), third value is the cosine similarity
        else:
            return oldExposures, originalSimilarity, cos_sim(genomes, np.dot(W,H))    #first value is the exprosure, second value is the similarity (based on the similarity matic), third value is the cosine similarity
    # The while loop starts here
    current_signatures=sub.get_items_from_index(signature_ids,np.nonzero(oldExposures)[0])
    while Flag: 
        
        # get the list of the indices those already have zero values
        if len(successList[1]) == 0:
            initialZerosIdx = list(np.where(oldExposures==0)[0]) 
            #get the indices to be selected
            selectableIdx = list(np.where(oldExposures>0)[0]) 
        elif len(successList[1]) > 1: 
            initialZerosIdx = list(np.where(successList[1]==0)[0]) 
            #get the indices to be selected
            selectableIdx = list(np.where(successList[1]>0)[0])
        else:
            print("iteration is completed")
            #break
        
        
        # get the total mutations for the given sample
        maxmutation = round(np.sum(genomes))
        
        # new signature matrix omiting the column for zero
        #Winit = np.delete(W, initialZerosIdx, 1)
        Winit = W[:,selectableIdx]
        
        # set the initial cos_similarity or other similarity distance
        record  = [np.inf, [], 1]   #
        # get the number of current nonzeros
        l= Winit.shape[1]
        
        background_sig = get_changed_background_sig_idx(list(oldExposures), background_sig)
        
        for i in range(l):
            
            if i in background_sig:
                continue
            
            loopSelection = list(range(l))
            del loopSelection[i]
            
            W1 = Winit[:,loopSelection]
           
            
            #initialize the guess
            x0 = np.random.rand(l-1, 1)*maxmutation
            x0= x0/np.sum(x0)*maxmutation
            
            #set the bounds and constraints
            bnds = create_bounds([], genomes, W1.shape[1]) 
            cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genomes]} 
            
            #the optimization step
            if solver == "slsqp":
                sol = minimize(parameterized_objective2_custom, x0, args=(W1, genomes),  bounds=bnds, constraints =cons1, tol=1e-15)
                
                #print (sol.success)
                #print (sol.x)
                
                #convert the newExposure vector into list type structure
                newExposure = list(sol.x)
            if solver == "nnls":
                ### using NNLS algorithm 
                reg = nnls(W1,genomes)
                weights = reg[0]
                newSample = np.dot(W1, weights)
                normalised_weights = weights/sum(weights)
                solution = normalised_weights*sum(genomes)                
                newExposure = list(solution)
                
            #insert the loopZeros in its actual position 
            newExposure.insert(i, 0)
            
            #insert zeros in the required position the newExposure matrix
            initialZerosIdx.sort()
            
            for zeros in initialZerosIdx:
                newExposure.insert(zeros, 0)
            
            # get the maximum value the new Exposure
            #maxcoef = max(newExposure)
            #idxmaxcoef = newExposure.index(maxcoef)
            
            #newExposure = np.round(newExposure)
            
            
            #if np.sum(newExposure)!=maxmutation:
                #newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
                
            newExposure=np.array(newExposure)
            
            if verbose==True:
                #print(newExposure)
                print("\nRemoving {}".format(current_signatures[i]))
            if metric == "cosine":
                newSimilarity = 1-cos_sim(genomes, newSample) 
                if verbose==True:
                    print("newSimilarity",1-newSimilarity) 
            elif metric == "l2":
                newSimilarity = np.linalg.norm(genomes-newSample, ord=2)/np.linalg.norm(genomes, ord=2)
                if verbose==True:
                    print("newSimilarity",newSimilarity) 
            difference =  newSimilarity-originalSimilarity 
            if difference<0:
                difference=cutoff+1e-100
            if verbose==True:
                print("difference", difference) 
            if difference<record[0]:
                record = [difference, newExposure, newSimilarity]
                
                #print("difference", difference)
            if verbose==True:
                print("--------------------------------------")
            
        if verbose==True:    
            print("\n############################\n############################") 
            #print("Selected Exposure")
            #print(record[1])
            selected_sigs=sub.get_items_from_index(signature_ids,np.nonzero(record[1])[0])
            dropped_sig=list(set(current_signatures)-set(selected_sigs))
            print("Dropped Signature: {}".format(dropped_sig))
            current_signatures=selected_sigs
            print("New Similarity", record[2])
            print("Similarity Difference: {}".format(record[2]-originalSimilarity))
            print("Current Signatures: {}".format(selected_sigs))
            print("****************************\n****************************\n\n")
            #print ("This loop's selection is {}".format(record))
        
        if record[0]>cutoff:   
            Flag=False
        elif len(record[1][np.nonzero(record[1])])==1:
            successList = record 
            Flag=False
        else:
            successList = record
            background_sig = get_changed_background_sig_idx(list(record[1]), background_sig)
            originalSimilarity=record[2]
        
        #print("The loop selection is {}".format(successList))
        
        #print (Flag)
        #print ("\n\n")
    
    #print ("The final selection is {}".format(successList))
    
    if len(successList[1])==0:
        successList = [0.0, oldExposures, originalSimilarity]
    
    if verbose==True:
        print("\n") 
        print("Final Exposure")
        print(successList[1])
        if metric=="cosine":
            print("Final Similarity: ", 1-successList[2])
        if metric=="l2":
            print("Final Similarity: ", successList[2])
    H = successList[1]
    return H, successList[2], cos_sim(genomes, np.dot(W,H)) #first value is the exprosure, second value is the similarity (based on the similarity matic), third value is the cosine similarity
    


def remove_all_single_signatures_pool(indices, W, exposures, totoalgenomes):
    i = indices
    H = exposures[:,i]
    genomes= totoalgenomes[:,i]
    # make the empty list of the successfull combinations
    successList = [0,[],0] 
    # get the cos_similarity with sample for the oringinal W and H[:,i]
    originalSimilarity= cos_sim(genomes, np.dot(W, H))
    # make the original exposures of specific sample round
    oldExposures = np.round(H)
    
    # set the flag for the while loop
    if len(oldExposures[np.nonzero(oldExposures)])>1:
        Flag = True
    else: 
        Flag = False
        return oldExposures
    # The while loop starts here
    while Flag: 
        
        # get the list of the indices those already have zero values
        if len(successList[1]) == 0:
            initialZerosIdx = list(np.where(oldExposures==0)[0]) 
            #get the indices to be selected
            selectableIdx = list(np.where(oldExposures>0)[0]) 
        elif len(successList[1]) > 1: 
            initialZerosIdx = list(np.where(successList[1]==0)[0]) 
            #get the indices to be selected
            selectableIdx = list(np.where(successList[1]>0)[0])
        else:
            print("iteration is completed")
            #break
        
        
        # get the total mutations for the given sample
        maxmutation = round(np.sum(genomes))
        
        # new signature matrix omiting the column for zero
        #Winit = np.delete(W, initialZerosIdx, 1)
        Winit = W[:,selectableIdx]
        
        # set the initial cos_similarity
        record  = [0.11, []] 
        # get the number of current nonzeros
        l= Winit.shape[1]
        
        for i in range(l):
            #print(i)
            loopSelection = list(range(l))
            del loopSelection[i]
            #print (loopSelection)
            W1 = Winit[:,loopSelection]
           
            
            #initialize the guess
            x0 = np.random.rand(l-1, 1)*maxmutation
            x0= x0/np.sum(x0)*maxmutation
            
            #set the bounds and constraints
            bnds = create_bounds([], genomes, W1.shape[1]) 
            cons1 ={'type': 'eq', 'fun': constraints1, 'args':[genomes]} 
            
            #the optimization step
            sol = minimize(parameterized_objective2_custom, x0, args=(W1, genomes),  bounds=bnds, constraints =cons1, tol=1e-15)
            
            #print (sol.success)
            #print (sol.x)
            
            #convert the newExposure vector into list type structure
            newExposure = list(sol.x)
            
            #insert the loopZeros in its actual position 
            newExposure.insert(i, 0)
            
            #insert zeros in the required position the newExposure matrix
            initialZerosIdx.sort()
            for zeros in initialZerosIdx:
                newExposure.insert(zeros, 0)
            
            # get the maximum value the new Exposure
            maxcoef = max(newExposure)
            idxmaxcoef = newExposure.index(maxcoef)
            
            newExposure = np.round(newExposure)
            
            
            if np.sum(newExposure)!=maxmutation:
                newExposure[idxmaxcoef] = round(newExposure[idxmaxcoef])+maxmutation-sum(newExposure)
                
            
            newSample = np.dot(W, newExposure)
            newSimilarity = cos_sim(genomes, newSample) 
             
            difference = originalSimilarity - newSimilarity
            #print(originalSimilarity)
            #print(newSample)
            #print(newExposure)
            #print(newSimilarity)
            
            #print(difference)
            #print (newExposure)
            #print (np.round(H))
            #print ("\n\n")
             
            if difference<record[0]:
                record = [difference, newExposure, newSimilarity]
            
            
        #print ("This loop's selection is {}".format(record))
        
        if record[0]>0.01:   
            Flag=False
        elif len(record[1][np.nonzero(record[1])])==1:
            successList = record 
            Flag=False
        else:
            successList = record
        #print("The loop selection is {}".format(successList))
        
        #print (Flag)
        #print ("\n\n")
    
    #print ("The final selection is {}".format(successList))
    
    if len(successList[1])==0:
        successList = [0.0, oldExposures, originalSimilarity]
    #print ("one sample completed")
    return successList[1]




def add_remove_signatures(W,
                          sample, 
                          metric="l2", 
                          solver="nnls", 
                          background_sigs = [], 
                          permanent_sigs = [], 
                          candidate_sigs="all", 
                          add_penalty = 0.05, 
                          remove_penalty=0.01,
                          check_rule_negatives =[],
                          checkrule_penalty = 1.00, 
                          allsigids = False, 
                          directory = os.getcwd()+"/Signature_assaignment_logfile.txt", 
                          connected_sigs=True,
                          verbose=False):
    
    lognote = open(directory, 'a')
 
    maxmutation=np.sum(np.array(sample))
    
    always_background = copy.deepcopy(permanent_sigs)
    M = sample   
    if candidate_sigs == "all":
        candidate_sigs = list(range(W.shape[1]))
    #first add each signatures
    original_distance = np.inf # a big number
    layer = 0
    
    # check the cosine_similarity with 4 signatures (highest)
    cosine_similarity_with_four_signatures = 1.0 # a value that will allow the signature not be reported as novel signature
    
    #set the signature's name 
    if type(allsigids) != bool:
        allsigids = sub.get_items_from_index(allsigids,candidate_sigs)
    else:
        allsigids = candidate_sigs
        
    while True:
        if verbose:
            print("\n\n\n\n!!!!!!!!!!!!!!!!!!STARTING LAYER: ", layer)
        lognote.write("\n\n!!!!!!!!!!!!!!!!!!!!!!!!! LAYER: {} !!!!!!!!!!!!!!!!!!!!!!!!!\n".format(layer))
        layer=layer+1
        layer_original_distance = np.inf  #a big number
        sigsToBeAdded = list(set(candidate_sigs)-set(background_sigs))
        #set the signature's name 
        if type(allsigids) != bool:
            allsigidsToBeAdded = sub.get_items_from_index(allsigids,sigsToBeAdded)
        else:
            allsigidsToBeAdded = sigsToBeAdded
        
        for i,j in zip(sigsToBeAdded, allsigidsToBeAdded):
            loop_sig = [i] 
            
            background_sigs = list(set(background_sigs).union(set(always_background)))
            if connected_sigs==True:
                background_sigs = add_connected_sigs(background_sigs, list(allsigids))
            
            add_exposures, add_distance, _ = add_signatures(W, M[:,np.newaxis], presentSignatures=copy.deepcopy(background_sigs), toBeAdded=loop_sig, 
                                                  metric="l2", verbose=False, check_rule_negatives=check_rule_negatives, check_rule_penalty=checkrule_penalty, cutoff=add_penalty) 
            
            if verbose:
                print("\n\n\n################## Add Index {} ########################".format(j)) 
                print(np.nonzero(add_exposures)[0])
                print(sigsToBeAdded)
                print(add_distance)
                
                
            remove_exposures, remove_distance, _ = remove_all_single_signatures(W, add_exposures, M, background_sigs = always_background, metric="l2", verbose = False, cutoff=remove_penalty)
            
            if verbose:
                print("\n################## Remove ########################")
                print(np.nonzero(remove_exposures)[0])
                print(remove_distance)
            # check if there is any change between the add and remove signatures and assign the distance and exposure accoridingly    
            if (np.nonzero(add_exposures)[0].all()==np.nonzero(remove_exposures)[0].all()) and np.nonzero(add_exposures)[0].shape == np.nonzero(remove_exposures)[0].shape:
                distance = add_distance
                exposures = add_exposures
            else:
                distance = remove_distance
                exposures = remove_exposures
            
            if distance < layer_original_distance:
                selected_signatures = np.nonzero(exposures)[0]
                layer_original_distance = distance 
                activities = exposures
        if verbose:
            print("\n\n#################### After Add-Remove #################################")
            print(selected_signatures)
            print(layer_original_distance)
        
        lognote.write("Best Signature Composition {}\n".format(sub.get_items_from_index(allsigids,selected_signatures)))    
        lognote.write("L2 Error % {}\n".format(round(layer_original_distance, 2))) 
        lognote.write("Cosine Similarity {}\n".format(round(cos_sim(M, np.dot(W,activities)),2)))         
        if layer_original_distance < original_distance:
            original_distance = layer_original_distance
            background_sigs = list(selected_signatures)
            finalactivities = activities
            if len(background_sigs)==4:
                cosine_similarity_with_four_signatures = cos_sim(M, np.dot(W,finalactivities))
                #print(cosine_similarity_with_four_signatures)
        else:
            N=np.dot(W,finalactivities)
            cosine_similarity = cos_sim(M, N)
            kldiv=round(scipy.stats.entropy(M,N),3)
            correlation,_=scipy.stats.pearsonr(M, N)
            correlation=round(correlation,2)
            break
    
    
    # get the maximum value of the new Exposure
    maxcoef = np.max(finalactivities)
    idxmaxcoef = list(finalactivities).index(maxcoef)
    
    finalactivities = np.round(finalactivities)
    
    # We may need to tweak the maximum value of the new exposure to keep the total number of mutation equal to the original mutations in a genome
    if np.sum(finalactivities)!=maxmutation:
        finalactivities[idxmaxcoef] = round(finalactivities[idxmaxcoef])+maxmutation-np.sum(finalactivities) 
        
    if verbose:
        print("\n########################## Final ###########################")        
        print(background_sigs)
        print(original_distance)
        print(finalactivities)
    lognote.write("\n#################### Final Composition #################################\n")
    lognote.write("{}\n".format(sub.get_items_from_index(allsigids,selected_signatures)))    
    lognote.write("L2 Error % {}\n".format(round(original_distance, 2))) 
    lognote.write("Cosine Similarity {}\n".format(round(cos_sim(M, np.dot(W,finalactivities)),2)))        
    #close lognote
    lognote.close()
    #newExposure, newSimilarity = fit_signatures(W[:,list(background_sigs)], M, metric="l2")
    #print(newExposure, newSimilarity)
    return (background_sigs, finalactivities, original_distance, cosine_similarity, kldiv, correlation, cosine_similarity_with_four_signatures)



import os
import tempfile
import shutil
import gzip
import contextlib
import logging
import functools

import numpy as np
import pandas as pd
import pysam

from handygenome.sigprofiler_clone.SigProfilerAssignment.single_sample import add_remove_signatures
from handygenome.sigprofiler_clone.SigProfilerMatrixGenerator.scripts.SigProfilerMatrixGeneratorFunc import SigProfilerMatrixGeneratorFunc

import handygenome.refgenome as refgenome
import handygenome.variant.varianthandler as varianthandler
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.signature.misc as signature_misc
import handygenome.signature.signatureresult as signatureresult
import handygenome.deco as deco


LOGGER = logging.getLogger(__name__)


def make_uncompressed_vcf_copy(vcf_path, input_copy_path):
    if vcf_path.endswith('.gz'):
        with \
                gzip.open(vcf_path, 'rb') as in_file, \
                open(input_copy_path, 'wb') as out_file:
            shutil.copyfileobj(in_file, out_file)
    else:
        os.symlink(vcf_path, input_copy_path)


@deco.get_deco_arg_choices({'refver': signature_misc.AVAILABLE_REFVERS})
def get_catalogues(vcf_path, refver, verbose=True):
    sampleid = 'test'
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmpdir:
        # run SigProfilerMatrixGeneratorFunc
        input_copy_path = os.path.join(tmpdir, 'input.vcf')
        make_uncompressed_vcf_copy(vcf_path, input_copy_path)
        matgenfunc_partial = functools.partial(
            SigProfilerMatrixGeneratorFunc,
            project=sampleid, genome=refver, vcfFiles=tmpdir,
            exome=False, bed_file=None, chrom_based=False, 
            plot=False, tsb_stat=False)

        if verbose:
            matgenfunc_partial()
        else:
            with open(os.devnull, 'w') as devnull:
                with \
                        contextlib.redirect_stdout(devnull), \
                        contextlib.redirect_stderr(devnull):
                    matgenfunc_partial()

        # load results
        catalogues = dict()
        for muttype in ('SBS', 'ID', 'DBS'):
            subdir = os.path.join(tmpdir, 'output', muttype)
            for fname in os.listdir(subdir):
                key = fname.split('.')[1]  # e.g. SBS96, ID83
                key_lower = key.lower()
                catalogues[key] = pd.read_csv(os.path.join(subdir, fname), 
                                              sep='\t', index_col=0)
                catalogues[key_lower] = catalogues[key]

    return catalogues


@deco.get_deco_arg_choices({'refver': signature_misc.AVAILABLE_REFVERS})
def get_catalogues_from_vcfspecs(vcfspec_iter, refver, verbose=True):
    with tempfile.TemporaryDirectory(dir=os.getcwd()) as tmpdir:
        # write a temporary vcf file
        input_vcf_path = os.path.join(tmpdir, 'input.vcf')
        chromdict = refgenome.ChromDict.from_refver(refver)
        header = initvcf.create_header(chromdict)
        with pysam.VariantFile(input_vcf_path, mode='w0', 
                               header=header) as out_vcf:
            for vcfspec in vcfspec_iter:
                vr = header.new_record()
                varianthandler.apply_vcfspec(vr, vcfspec)
                out_vcf.write(vr)
        # create catalogues 
        catalogues = get_catalogues(input_vcf_path, refver, verbose=verbose)

    return catalogues


#############################################


def turn_signames_into_indexes(signames, sigdata):
    result = list()
    sigdata_signames = tuple(sigdata.columns)
    for x in signames:
        if x in sigdata_signames:
            result.append(sigdata_signames.index(x))
        else:
            LOGGER.warning(f'Function "run_assignment": '
                           f'Signature name {x} is not included in the '
                           f'input signature dataset.')

    return result


def add_recommended_sigs(sigdata, background_sigs, permanent_sigs):
    recommended_sigs = ['SBS1', 'SBS5']
    if set(recommended_sigs).issubset(set(sigdata.columns)):
        background_sigs.extend(recommended_sigs)
        permanent_sigs.extend(recommended_sigs)


def run_assignment_sanity_check(catalogue, sigdata):
    if set(catalogue.index) != set(sigdata.index):
        raise Exception(f'The mutation catalogue type of "catalogue" and '
                        f'"sigdata" are different.')


def reorder_catalogue(catalogue, sigdata):
    if tuple(catalogue.index) != tuple(sigdata.index):
        catalogue_reorderd = catalogue.reindex(sigdata.index)
        return catalogue_reordered
    else:
        return catalogue


def run_assignment(catalogue: pd.Series,
                   sigdata: pd.DataFrame,
                   background_sigs=list(), 
                   permanent_sigs=list(),
                   checkrule_sigs=list(), 
                   add_penalty=0.05, 
                   remove_penalty=0.01,
                   checkrule_penalty=1.00,
                   use_recommended_sigs=True,
                   use_connected_sigs=True, 
                   verbose=False):
    """Args:
        catalogue: pandas.Series representing the mutation catalogue of 
            the sample. Its index must indicate mutation contexts 
            (e.g. C[C>T]A) and each value must represent the mutation count.
        sigdata: pandas.DataFrame where rows indicate mutation contexts 
            (e.g. C[C>T]A) and columns indicate signature components.

            If must be that "set(catalogue.index) == set(sigdata.index)".
            If the order is different, "catalogue" is reordered by 
            "pd.Series.reindex" method.

        background_sigs: These signatures are included at the beginning of
            the add-remove cycles.
        permanent_sigs: These signatures are included at the beginning of
            the add-remove cycles. In addition, these are not removed during
            remove cycles.
        add_penalty: A signature component can be added if similarity gain is
            at least this value.
        checkrule_sigs: These signatures get penalty during addition cycles.
        checkrule_penalty: This value is multiplied to the raw similarity
            during checkrule penalty application.
        use_connected_sigs: If True, "background_sigs" is augmented with related 
            signatures as follows:
                ["SBS2", "SBS13"]
                ["SBS7a", "SBS7b", "SBS7c", "SBS7d"]
                ["SBS10a", "SBS10b"]
                ["SBS17a", "SBS17b"]

    Raises:
        When set(catalogue.index) != set(sigdata.index)
    """
    # sanity check
    run_assignment_sanity_check(catalogue, sigdata)
    # reorder "catalogue" if its catalogue key order is different with sigdata
    catalogue = reorder_catalogue(catalogue, sigdata)
    # apply recommended background_sigs and permanent_sigs
    if use_recommended_sigs:
        add_recommended_sigs(sigdata, background_sigs, permanent_sigs)
    # make a temporary file for logging
    fd, logfile_path = tempfile.mkstemp(dir=os.getcwd(), 
                                        prefix='SigProfiler_log.', 
                                        suffix='.txt')
    os.close(fd)
    # convert pd.DataFrame into np.ndarray
    catalogue_array = np.array(catalogue)
    sigdata_array = np.array(sigdata)
    # convert arguments
    background_sigs_indexes = turn_signames_into_indexes(background_sigs,
                                                         sigdata)
    permanent_sigs_indexes = turn_signames_into_indexes(permanent_sigs,
                                                        sigdata)
    checkrule_sigs_indexes = turn_signames_into_indexes(checkrule_sigs,
                                                        sigdata)
    # run main function
    (background_sigs, 
     finalactivities, 
     original_distance, 
     cosine_similarity, 
     kldiv, 
     correlation, 
     cosine_similarity_with_four_signatures) = add_remove_signatures(
        W=sigdata_array, sample=catalogue_array,
        background_sigs=background_sigs_indexes,
        permanent_sigs=permanent_sigs_indexes,
        add_penalty=add_penalty, remove_penalty=remove_penalty,
        check_rule_negatives=checkrule_sigs_indexes,
        checkrule_penalty=checkrule_penalty,
        connected_sigs=use_connected_sigs, verbose=verbose)

    # remove temporary log file
    os.remove(logfile_path)

    # create sigresult
    exposure = pd.Series(dict(zip(sigdata.columns, finalactivities)))
    sigresult = signatureresult.SignatureResult(catalogue=catalogue,
                                                sigdata=sigdata,
                                                exposure=exposure,
                                                cossim=cosine_similarity,
                                                kldiv=kldiv,
                                                correlation=correlation)

    return sigresult






# svcparser: SV caller parser
import re
#import warnings

import pysam

import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))
libbreakends = importlib.import_module('.'.join([top_package_name, 'sv', 'breakends']))


INFO_END_PAT = re.compile('^(.*;)?END=([^;]+);.*$')


def get_info_end(vr):
    candidates = list()

    infosplit = str(vr).split('\t')[7].strip().split(';')
    for item in infosplit:
        itemsp = item.split('=')
        if itemsp[0] == 'END':
            candidates.append(int(itemsp[1]))
            break

    if len(candidates) == 0:
        raise Exception(f'INFO does not include END in the input variant '
                        f'record:\n{vr}')

    return candidates[0]


#######################################################


def get_bnds_from_caller_vr(vr, fasta, chromdict):
    """
    Raises:
        All exceptions are from get_vr_svinfo_caller_vr function.
    """

    if len(vr.alts) != 1:
        raise Exception(
            f'Unexpected caller output vcf pattern: '
            f'Multialleleic SV variant record:\n{vr}')

    vr_svinfo = get_vr_svinfo_caller_vr(vr, fasta, chromdict) 
        # this function does not return None ; it rather raises an Exception
    if vr_svinfo is None:
        return None
    else:
        bnds = libbreakends.get_bnds_from_vr_svinfo(vr, vr_svinfo, fasta, 
                                                        chromdict)
        bnd1adv_form = libbreakends.get_bnds_equivalents(bnds)[0]
        return bnd1adv_form


def get_vr_svinfo_caller_vr(vr, fasta, chromdict):
    """
    Raises:
        If input vr does not conform to a known SV variant record format 
            (including a valid non-SV variant record)
    """

    if check_vr_format_delly(vr):
        vr_svinfo = get_vr_svinfo_delly(vr, fasta, chromdict)

    elif check_vr_format_manta(vr):
        vr_svinfo = get_vr_svinfo_manta(vr, fasta, chromdict)

    else: # Assumed that other SV callers follow the SV vcf standard
        vr_svinfo = libbreakends.get_vr_svinfo_standard_vr(vr, fasta, 
                                                               chromdict)
        modify_vr_svinfo_for_dRanger(vr_svinfo, fasta)

    return vr_svinfo


########################################################


def modify_vr_svinfo_for_dRanger(vr_svinfo, fasta):
    if vr_svinfo['chrom_mate'] not in fasta.references:
        assert vr_svinfo['chrom_mate'].startswith('chr')
        vr_svinfo['chrom_mate'] = re.sub('^chr', '', vr_svinfo['chrom_mate'])


def get_vr_svinfo_manta(vr, fasta, chromdict):
    if vr.alts[0].startswith('<') and vr.alts[0].endswith('>'): # non-TRA
        vr_svinfo = dict()

        vr_svinfo['chrom_mate'] = vr.contig
        vr_svinfo['pos_mate'] = get_info_end(vr)

        assert vr_svinfo['pos_mate'] > vr.pos, (
            f'INFO/END is smaller than POS for this presumed MANTA non-BND '
            f'variant record:\n{vr}')
        assert vr.alts[0] in ('<DUP:TANDEM>', '<DEL>'), (
            f'Unexpected ALT string for this presumed MANTA non-BND '
            f'variant record:\n{vr}')

        if vr.alts[0] == '<DUP:TANDEM>':
            vr_svinfo['current_endis5'] = True
            vr_svinfo['mate_endis5'] = False
        elif vr.alts[0] == '<DEL>':
            vr_svinfo['current_endis5'] = False
            vr_svinfo['mate_endis5'] = True

        vr_svinfo['ref'] = fasta.fetch(vr.contig, vr.pos - 1, vr.pos)
        vr_svinfo['t'] = vr_svinfo['ref']

        vr_svinfo['is_bnd1'] = True
    
    else: # standard bnd string representation is assumed
        vr_svinfo = libbreakends.get_vr_svinfo_standard_vr(vr, fasta, 
                                                               chromdict)

    return vr_svinfo


def get_vr_svinfo_delly(vr, fasta, chromdict):
    #assert vr.ref == 'N', f'REF is not "N" for the input delly variant record:\n{vr}'

    if vr.alts[0] == '<INS>':
        #warnings.warn(f'ALT is <INS> for input delly variant record.')
        return None
    else:
        vr_svinfo = dict()

        vr_svinfo['chrom_mate'] = vr.info['CHR2']
        #vr_svinfo['pos_mate'] = vr.stop 
        vr_svinfo['pos_mate'] = get_info_end(vr)
            # delly INFO/END value seems to be 1-based
            # 220323: vr.stop does not return INFO/END value for some delly variant records

        CTsplit = vr.info['CT'].split('to')
        vr_svinfo['current_endis5'] = (CTsplit[0] == '5')
        vr_svinfo['mate_endis5'] = (CTsplit[1] == '5')

        vr_svinfo['ref'] = fasta.fetch(vr.contig, vr.pos - 1, vr.pos)
        vr_svinfo['t'] = vr_svinfo['ref']

        vr_svinfo['is_bnd1'] = libbreakends.get_is_bnd1(vr, vr_svinfo, chromdict)

        return vr_svinfo


######################################################################################


def check_vr_format_delly(vr):
    return (('CHR2' in vr.info) and 
            ('CHR1' not in vr.info) and 
            ('CT' in vr.info) and 
            (vr.ref == 'N'))


def check_vr_format_manta(vr):
    # For BND records, ALT is VCF breakends string
    # For non-BND, ALT is like <DUP:TANDEM>
    # ID starts with 'MANTA'

    if ( vr.id is not None and vr.id.startswith('Manta') ):
        return True
    else:
        return False


def check_vr_format_brass(vr):
    existing_infokey_set = set(('TRDS'))
    if existing_infokey_set.issubset(set(vr.info.keys())):
        return True
    else:
        return False


def check_vr_format_dRanger(vr):
    existing_infokey_set = set(('SVCLASS', 'DRQUAL'))
    if existing_infokey_set.issubset(set(vr.info.keys())):
        return True
    else:
        return False


def check_vr_format_snowman(vr):
    existing_infokey_set = set(('EVDNC', 'SCTG'))
    if existing_infokey_set.issubset(set(vr.info.keys())):
        return True
    else:
        return False



# Only deals with command-line VEP
import os
import subprocess

import pysam

import handygenome
import handygenome.refgenome as refgenome
import handygenome.workflow as workflow
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.variant.vcfspec as libvcfspec


VEP_INFO_FIELD = 'CSQ'
DEFAULT_DISTANCE = 5000
REFVER_TO_VEPARGS = {
    'hg19' : {'species': 'homo_sapiens', 'assembly': 'GRCh37'},
    'hg38' : {'species': 'homo_sapiens', 'assembly': 'GRCh38'},
    'mm10' : {'species': 'mus_musculus', 'assembly': 'GRCm38'},
    'mm39' : {'species': 'mus_musculus', 'assembly': 'GRCm39'},
}

ALLOWED_SPECIES_ASSEMBLIES = {
    'homo_sapiens': ('GRCh37', 'GRCh38'),
    'mus_musculus': ('GRCm38', 'GRCm39'),
    }


def check_species_assembly_sanity(species, assembly):
    if species not in ALLOWED_SPECIES_ASSEMBLIES.keys():
        raise Exception(f'"species" must be one of '
                        f'{ALLOWED_SPECIES_ASSEMBLIES.keys()}.')
    else:
        if assembly not in ALLOWED_SPECIES_ASSEMBLIES[species]:
            raise Exception(
                f'For the species {species}, assembly name must be one of '
                f'{ALLOWED_SPECIES_ASSEMBLIES[species]}.')


def get_vep_args(infile_path, outfile_path, fasta_path, species, assembly,
                 distance=DEFAULT_DISTANCE, force_overwrite=True):
    check_species_assembly_sanity(species, assembly)

    if species == 'mus_musculus' and assembly == 'GRCm38':
        #vep_path = common.VEP_MM10
        vep_path = handygenome.OPTION['vep_mm10']
    else:
        #vep_path = common.VEP
        vep_path = handygenome.OPTION['vep']

    with pysam.FastaFile(fasta_path) as fasta:
        max_chrom_size = max(fasta.lengths)

    vep_args = [
        vep_path,
        '-i', f'{infile_path}',
        '-o', f'{outfile_path}',
        '--fasta', f'{fasta_path}',
        '--species', f'{species}',
        '--assembly', f'{assembly}',
        '--max_sv_size', f'{max_chrom_size}',

        #'--dir', f'{common.VEP_CACHE_DIR}',
        '--dir', handygenome.OPTION['vep_cache_dir'],
        '--cache',
        '--offline',

        '--vcf',
        '--vcf_info_field', VEP_INFO_FIELD,
        '--terms', 'SO',
        '--shift_hgvs', '0',
        '--dont_skip',
        '--no_stats',

        '--distance', f'{distance}',

        '--symbol',
        '--numbers',
        '--hgvsg',
        '--hgvs',
        '--ccds',
        #'--mirna',
        '--biotype',
        '--canonical',
        '--xref_refseq',
        '--mane',

        '--regulatory',

        '--protein',
        '--uniprot',
        '--sift', 'b',
        '--polyphen', 'b',

        #'--check_existing',
        #'--pubmed',
        #'--af',

        #'--max_af',
        #'--af_1kg',
        #'--af_esp',
        #'--af_gnomad',
        #'--var_synonyms',

        #'--nearest', 'transcript',
    ]

    if force_overwrite:
        vep_args.append('--force_overwrite')

    return vep_args


def run_vep_with_vcfspec(vcfspec, refver, distance=DEFAULT_DISTANCE):
    chromdict = refgenome.ChromDict(refver = refver)
    vr = initvcf.create_vr(chromdict=chromdict, vcfspec=vcfspec)

    return run_vep_with_vr(vr, refver, distance)


def run_vep_with_interval(interval, refver, distance=DEFAULT_DISTANCE):
    chromdict = refgenome.ChromDict(refver = refver)
    vcfspec = libvcfspec.Vcfspec(interval.chrom, interval.start0, 'N', ['<DEL>'])
    while True:
        vr = initvcf.create_vr(chromdict=chromdict, vcfspec=vcfspec, 
                               end=interval.end1)
        if vr.stop == interval.end1:
            break
        else:
            print(f'INFO/END of created VariantRecord object is faulty. '
                  f'Retrying.')
            continue

    return run_vep_with_vr(vr, refver, distance)


def run_vep_with_vr(vr, refver=None, distance=DEFAULT_DISTANCE):
    if refver is None:
        refver = refgenome.infer_refver_vr(vr)

    assert refver in REFVER_TO_VEPARGS.keys()

    species = REFVER_TO_VEPARGS[refver]['species']
    assembly = REFVER_TO_VEPARGS[refver]['assembly']

    fasta_path = refgenome.get_default_fasta_path(refver)
    infile_path = workflow.get_tmpfile_path(delete=True)
    outfile_path = infile_path + '.vep'

    with pysam.VariantFile(infile_path, mode='wz', 
                           header=vr.header) as out_vcf:
        out_vcf.write(vr)

    vep_args = get_vep_args(infile_path, outfile_path, fasta_path, species, 
                            assembly, distance=distance)

    p = subprocess.run(args=vep_args, capture_output=True, text=True,
                       check=True)

    with pysam.VariantFile(outfile_path) as in_vcf:
        out_vr = next(in_vcf.fetch())

    os.remove(infile_path)
    os.remove(outfile_path)

    return out_vr





import pysam



class CNVBrowser:
    def __init__(self, refver, normal_bam_path=None, tumor_bam_path=None, region_gr=None):
        self.refver = refver
        self.region_gr = region_gr
        self.normal_bam_path = normal_bam_path
        self.tumor_bam_path = tumor_bam_path
        if self.normal_bam_path is None:
            self.normal_bam = pysam.AlignmentFile(self.normal_bam_path)
        if self.tumor_bam_path is None:
            self.tumor_bam = pysam.AlignmentFile(self.tumor_bam_path)

    def __del__(self):
        for bam in (self.normal_bam, self.tumor_bam):
            if bam is not None:
                bam.close()

    def set_gcdata(self):
        pass



import os
import re
import argparse
import gzip
import ftplib
import contextlib

import pysam

import handygenome.network as network
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.refgenome as refgenome
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.vcfeditor.indexing as indexing
import handygenome.publicdb.ncbi as libncbi


DEFAULT_OUTFILE_BASENAME_GRCH37 = "dbSNP_b155_GRCh37.p13.vcf.gz"
DEFAULT_OUTFILE_BASENAME_GRCH38 = "dbSNP_b155_GRCh38.p13.vcf.gz"
DEFAULT_OUTFILE_PATH_GRCH37 = os.path.join(os.getcwd(), DEFAULT_OUTFILE_BASENAME_GRCH37)
DEFAULT_OUTFILE_PATH_GRCH38 = os.path.join(os.getcwd(), DEFAULT_OUTFILE_BASENAME_GRCH38)

#URL_GRCH37 = "https://ftp.ncbi.nih.gov/snp/archive/b155/VCF/GCF_000001405.25.gz"
#URL_GRCH37_MD5 = "https://ftp.ncbi.nih.gov/snp/archive/b155/VCF/GCF_000001405.25.gz.md5"
#URL_GRCH38 = "https://ftp.ncbi.nih.gov/snp/archive/b155/VCF/GCF_000001405.39.gz"
#URL_GRCH38_MD5 = "https://ftp.ncbi.nih.gov/snp/archive/b155/VCF/GCF_000001405.39.gz.md5"

URL_GRCH37 = "https://ftp.ncbi.nih.gov/snp/latest_release/VCF/GCF_000001405.25.gz"
URL_GRCH37_MD5 = "https://ftp.ncbi.nih.gov/snp/latest_release/VCF/GCF_000001405.25.gz.md5"
URL_GRCH38 = "https://ftp.ncbi.nih.gov/snp/latest_release/VCF/GCF_000001405.40.gz"
URL_GRCH38_MD5 = "https://ftp.ncbi.nih.gov/snp/latest_release/VCF/GCF_000001405.40.gz.md5"


def argument_parser(cmdargs):
    parser_dict = workflow.init_parser(
        description=(
            f"The original dbSNP VCF file is downloaded, then "
            f"modified to a handygenome-compatible VCF file."
        )
    )

    parser_dict["required"].add_argument(
        "--outfile",
        dest="outfile_path",
        #default=DEFAULT_OUTFILE_PATH_GRCH37,
        required=True,
        help=f"Output vcf file path",
    )

    parser_dict["optional"].add_argument(
        "--downloaded-file",
        dest="download_path",
        required=False,
        help=f"Previously downloaded dbSNP file path",
    )

    workflow.add_refver_arg(
        parser_dict['required'], required=True, choices=['GRCh37', 'GRCh38'],
    )
    workflow.add_logging_args(parser_dict)

    args = parser_dict["main"].parse_args(cmdargs)

    return args


# Downloading original VCF file

def get_refver(url):
    response = network.http_run_urlopen(url)
    gf = gzip.GzipFile(fileobj=response, mode='rb')
    for bline in gf:
        if bline.startswith(b'##reference'):
            refver = bline.decode().split('=')[1]  # GRCh37.p13
            refver = refver.split('.')[0]
            return refver
        if not bline.startswith(b'##'):
            break

    raise Exception(f'No reference version information found from VCF header')


def get_refver_pysam(url):
    with pysam.VariantFile(url) as vcf:
        for record in vcf.header.records:
            if record.key == 'reference':
                refver = record.value  # GRCh37.p13
                refver = refver.split('.')[0]
                return refver

    raise Exception(f'No reference version information found from VCF header')


def pick_vcf_url(url_dict, is_grch37=True):
    if is_grch37:
        key = min(url_dict.keys(), key=(lambda x: int(x.split('.')[1])))
    else:
        key = max(url_dict.keys(), key=(lambda x: int(x.split('.')[1])))

    url = url_dict[key]['vcf']
    refver = get_refver(url)
    #refver = get_refver_pysam(url)
        # not used due to connection failure risk
    if (
        (is_grch37 and (refver != 'GRCh37'))
        or
        ((not is_grch37) and (refver != 'GRCh38'))
    ):
        raise Exception(f'Unexpected filename-refver pairing; url_dict={url_dict}, url={url}, refver={refver}')

    return url


def download(logger, url, download_path):
    logger.info(f'Beginning download of url "{url}" to "{download_path}"')
    network.download(url, download_path)
    logger.info(f"Download finished")


##########################################

# Modification of original VCF file

def get_generic_metadata(vcf):
    result = dict()
    for rec in vcf.header.records:
        if rec.type == 'GENERIC':
            result[rec.key] = rec.value

    return result


def get_dbsnp_build(vcf):
    for record in vcf.header.records:
        if record.key == 'dbSNP_BUILD_ID':
            build = record.value  # class 'str'
            return build

    raise Exception(f'No build information found from VCF header')


def get_freq_classes(vcf, logger):
    result = set()
    NR = 0
    for vr in vcf.fetch():
        NR += 1
        if NR % 10_000_000 == 0:
            logger.info(
                f"get_freq_classes: Processing {NR:,}th variant " f"record:\n{vr}"
            )

        try:
            freq = vr.info["FREQ"]
        except KeyError:
            continue

        freq = ",".join(freq)
        classes = [x.split(":")[0] for x in freq.split("|")]
        result.update(classes)

    result = sorted(result)

    return result


def get_new_header(chromdict, freq_classes, original_vcf):
    new_header = initvcf.create_header(chromdict=chromdict)

    for rec in original_vcf.header.records:
        if rec.key in ('fileDate', 'dbSNP_BUILD_ID', 'reference'):
            new_header.add_record(rec)

    new_header.add_meta(
        key="INFO",
        items=[
            ("ID", "rs"),
            ("Type", "Integer"),
            ("Number", 1),
            ("Description", "dbSNP rs ID"),
        ],
    )
    new_header.add_meta(
        key="INFO",
        items=[
            ("ID", "dbSNPBuildID"),
            ("Type", "Integer"),
            ("Number", 1),
            ("Description", "First dbSNP Build for RS"),
        ],
    )
    new_header.add_meta(
        key="INFO",
        items=[
            ("ID", "COMMON"),
            ("Type", "Flag"),
            ("Number", 0),
            (
                "Description",
                (
                    "RS is a common SNP. A common SNP is one that has at least "
                    "one 1000Genomes population with a minor allele of "
                    "frequency >= 1% and for which 2 or more founders "
                    "contribute to that minor allele frequency."
                ),
            ),
        ],
    )

    for pop in freq_classes:
        new_header.add_meta(
            key="INFO",
            items=[
                ("ID", f"AF_{pop}"),
                ("Type", "Float"),
                ("Number", 1),
                ("Description", f"Allele frequency in the population '{pop}'"),
            ],
        )

    return new_header


def handle_freq(freq):
    freq_split = ",".join(freq).split("|")
    freq_dict = dict()
    for item in freq_split:
        item_split = item.split(":")
        freq_dict[item_split[0]] = [
            (float(0) if x == "." else float(x)) for x in item_split[1].split(",")
        ]

    return freq_dict


def get_new_vr_list(
    vr, chromdict, chrname_converter, new_header, freq_classes, infile_vcfspecs
):
    new_vr_list = list()
    infile_vcfspecs.setdefault(vr.contig, set())

    new_chrom = chrname_converter[vr.contig]
    if (
        (new_chrom is not None)
        and (new_chrom in chromdict.contigs)
        and "FREQ" in vr.info
    ):
        freq = vr.info["FREQ"]
        freq_dict = handle_freq(freq)
        for idx, alt in enumerate(vr.alts):
            vcfspec = (vr.pos, vr.ref, alt)
            if vcfspec in infile_vcfspecs[vr.contig]:
                continue

            infile_vcfspecs[vr.contig].add(vcfspec)

            new_vr = new_header.new_record()
            new_vr.chrom = new_chrom
            new_vr.pos = vr.pos
            new_vr.ref = vr.ref
            new_vr.alts = (alt,)

            new_vr.info["rs"] = vr.info["RS"]
            if "dbSNPBuildID" in vr.info:
                new_vr.info["dbSNPBuildID"] = vr.info["dbSNPBuildID"]
            if vr.info["COMMON"]:
                new_vr.info["COMMON"] = True

            for pop in freq_classes:
                if pop in freq_dict:
                    AF = freq_dict[pop][idx + 1]
                    new_vr.info[f"AF_{pop}"] = AF

            new_vr_list.append(new_vr)

    return new_vr_list


def write_outfile(
    original_vcf,
    outfile_path,
    chromdict,
    chrname_converter,
    new_header,
    freq_classes,
    logger,
):
    infile_vcfspecs = dict()
    with pysam.VariantFile(outfile_path, mode="wz", header=new_header) as out_vcf:
        for idx, vr in enumerate(original_vcf.fetch()):
            if idx % 1_000_000 == 0:
                logger.info(
                    f"write_outfile: Processing {idx + 1:,}th variant record:\n{vr}"
                )

            new_vr_list = get_new_vr_list(
                vr,
                chromdict,
                chrname_converter,
                new_header,
                freq_classes,
                infile_vcfspecs,
            )
            for new_vr in new_vr_list:
                out_vcf.write(new_vr)


#def write_outfile_old(
#    download_path,
#    outfile_path,
#    chromdict,
#    chrname_converter,
#    new_header,
#    freq_classes,
#    logger,
#):
#    infile_vcfspecs = dict()
#    with pysam.VariantFile(
#        outfile_path, mode="wz", header=new_header
#    ) as out_vcf, pysam.VariantFile(download_path, mode="r") as in_vcf:
#        NR = 0
#        for vr in in_vcf.fetch():
#            NR += 1
#            if NR % 1_000_000 == 0:
#                logger.info(
#                    f"write_outfile: Processing {NR:,}th variant " f"record:\n{vr}"
#                )
#
#            new_vr_list = get_new_vr_list(
#                vr,
#                chromdict,
#                chrname_converter,
#                new_header,
#                freq_classes,
#                infile_vcfspecs,
#            )
#            for new_vr in new_vr_list:
#                out_vcf.write(new_vr)


#def main_old(
#    download_path,
#    download_url,
#    outfile_path,
#    chromdict,
#    chrname_converter,
#    refver,
#    build,
#    logger,
#):
#    # download
#    if download_path is None:
#        download_path = workflow.get_tmpfile_path()
#        download(logger, download_url, download_path)
#        rm_downloaded = True
#    else:
#        rm_downloaded = False
#
#    # collect FREQ population names
#    logger.info("Collecting FREQ population names from the original file")
#    freq_classes = get_freq_classes(download_path, logger)
#
#    # write output file
#    new_header = get_new_header(chromdict, freq_classes, refver=refver, build=build)
#    logger.info("Modifying and writing original variant records")
#    write_outfile(
#        download_path,
#        outfile_path,
#        chromdict,
#        chrname_converter,
#        new_header,
#        freq_classes,
#        logger,
#    )
#
#    # indexing
#    indexing.index_vcf(outfile_path, overwrite=True)
#
#    # remove temp files
#    if rm_downloaded:
#        os.remove(download_path)


def main(
    download_path,
    outfile_path,
    is_grch37,
    logger,
):
    # download
    if download_path is None:
        url_dict = libncbi.get_dbsnp_urls()
        ftp_url = pick_vcf_url(url_dict, is_grch37=is_grch37)

        download_path = workflow.get_tmpfile_path(prefix="dbSNP_vcf_download_tmpfile_", suffix=".vcf.gz", dir=os.getcwd(), is_dir=False, delete=False)
        logger.info(f"Downloading original VCF file from NCBI ftp server to {repr(download_path)}")
        network.download_wget(ftp_url, download_path)
        logger.info("Download finished")

        url = download_path
        rm_tmp = True
    else:
        url = download_path
        rm_tmp = False

    original_vcf = pysam.VariantFile(url)

    # collect FREQ population names
    logger.info("Collecting FREQ population names from the original file")
    freq_classes = get_freq_classes(original_vcf, logger)
    logger.info("Finished collecting FREQ population names")

    # write output file
    if is_grch37:
        assembly_spec = refgenome.SPECS["GRCh37"]
        chromdict = assembly_spec.chromdicts["nochr_plus_genbank"]
        chrname_converter = assembly_spec.dicts[("refseq", "nochr_plus_genbank")]
    else:
        assembly_spec = refgenome.SPECS["GRCh38"]
        chromdict = assembly_spec.chromdicts["ucsc"]
        chrname_converter = assembly_spec.dicts[("refseq", "ucsc")]

    new_header = get_new_header(chromdict, freq_classes, original_vcf)
    logger.info("Modifying original variant records and writing to the output file")
    write_outfile(
        original_vcf,
        outfile_path,
        chromdict,
        chrname_converter,
        new_header,
        freq_classes,
        logger,
    )
    logger.info("Finished writing the output file")

    # indexing
    indexing.index_vcf(outfile_path, overwrite=True)

    # remove downloaded file
    if rm_tmp:
        os.remove(download_path)


#def main_grch37(args, logger):
#    asmblspec = refgenome.SPECS["grch37"]
#    main_old(
#        download_path=args.download_path_grch37,
#        download_url=URL_GRCH37,
#        outfile_path=args.outfile_path_grch37,
#        chromdict=asmblspec.chromdicts["nochr_plus_genbank"],
#        chrname_converter=asmblspec.dicts[("refseq", "nochr_plus_genbank")],
#        refver="GRCh37.p13",
#        build="155",
#        logger=logger,
#    )
#
#
#def main_grch38(args, logger):
#    asmblspec = refgenome.SPECS["grch38"]
#    main_old(
#        download_path=args.download_path_grch38,
#        download_url=URL_GRCH38,
#        outfile_path=args.outfile_path_grch38,
#        chromdict=asmblspec.chromdicts["ucsc"],
#        chrname_converter=asmblspec.dicts[("refseq", "ucsc")],
#        refver="GRCh38.p13",
#        build="155",
#        logger=logger,
#    )




import itertools
import warnings
import operator
import multiprocessing

import numpy as np
import pandas as pd
import pyranges as pr

import handygenome.refgenome as refgenome
import handygenome.tools as tools
import handygenome.cnv.misc as cnvmisc
import handygenome.logutils as logutils


#import handygenome.workflow as workflow
#LOGGER_INFO = workflow.get_debugging_logger(verbose=False)
#LOGGER_DEBUG = workflow.get_debugging_logger(verbose=True)


def sort_df_by_coord(df, chromdict):
    coord_cols = df.loc[:, ['Chromosome', 'Start', 'End']]
    sorted_rows = sorted(
        coord_cols.iterrows(),
        key=(lambda x: (chromdict.contigs.index(x[1][0]), x[1][1], x[1][2])),
    )
    sorted_rows_indexes = [x[0] for x in sorted_rows]
    return df.iloc[sorted_rows_indexes, :].reset_index(drop=True)


def isec_union(gr1, gr2, as_gr=True):
    gr1 = cnvmisc.arg_into_gr(gr1).drop()
    gr2 = cnvmisc.arg_into_gr(gr2).drop()

    isec = gr1.intersect(gr2)
    gr1_diff_isec = gr1.subtract(isec)
    gr2_diff_isec = gr2.subtract(isec)
    result = pr.concat([gr1_diff_isec, isec, gr2_diff_isec])
    result = result.sort()
    if as_gr:
        return result
    else:
        return result.df


# join #


#def inner_join(left_gr, right_gr):
#    """For handling grs with lots of columns"""
#    new_right_df = right_gr.df.loc[:, ['Chromosome', 'Start', 'End']]
#    new_right_df.insert(3, 'index', range(new_right_df.shape[0]))
#    new_right_gr = pr.PyRanges(new_right_df)
#    joined_gr = left_gr.join(new_right_gr)
#    right_values_df = right_gr.df.iloc[joined_gr.index.to_list(), 3:].reset_index(drop=True)
#    right_values_df.reset_index(drop=True, inplace=True)
#    
#    old_columns = left_gr.columns.to_list()
#    new_columns = [
#        (f'{x}_b' if x in old_columns else x)
#        for x in right_values_df.columns
#    ]
#    right_values_df.columns = new_columns
#
#    result_df = pd.concat([joined_gr.df.iloc[:, :-1], right_values_df], axis=1)
#    return pr.PyRanges(result_df)
#
#
#def add_new_coordinates(joined_gr, how):
#    """Accepts a result of 'join' method, which has columns 'Start_b' and 'End_b'"""
#    assert how in ("union", "intersection", "swap")
#    assert {'Start_b', 'End_b'}.issubset(joined_gr.columns)
#
#    Start = joined_gr.Start
#    End = joined_gr.End
#
#    _ = joined_gr.new_position(how)
#    new_Start = joined_gr.Start
#    new_End = joined_gr.End
#
#    joined_gr.Start = Start
#    joined_gr.End = End
#    joined_gr.new_Start = new_Start
#    joined_gr.new_End = new_End
#
#    return joined_gr
#
#
## this is of similar speed to add_new_coordinates
#def add_new_coordinates_new(joined_gr, how):
#    """Accepts a result of 'join' method, which has columns 'Start_b' and 'End_b'"""
#    assert how in ("union", "intersection")
#    assert {'Start_b', 'End_b'}.issubset(joined_gr.columns)
#
#    df = joined_gr.df
#    starts_subarr = df.loc[:, ['Start', 'Start_b']].to_numpy()
#    ends_subarr = df.loc[:, ['End', 'End_b']].to_numpy()
#
#    if how == 'union':
#        new_starts = starts_subarr.min(axis=1)
#        new_ends = ends_subarr.max(axis=1)
#    elif how == 'intersection':
#        new_starts = starts_subarr.max(axis=1)
#        new_ends = ends_subarr.min(axis=1)
#
#    joined_gr.new_Start = new_starts
#    joined_gr.new_End = new_ends
#
#    return joined_gr
#
#
## join helper
#def extract_unmatched_rows(left_gr, right_gr, joined_gr, find_nearest, added_columns):
#    unmatched_rows_gr = left_gr.subtract(joined_gr)
#    if unmatched_rows_gr.empty:
#        return None
#
#    if find_nearest:
#        unmatched_rows_gr = unmatched_rows_gr.nearest(
#            right_gr, overlap=False, how=None,
#        )
#    else:
#        nrow = unmatched_rows_gr.df.shape[0]
#        for colname in added_columns:
#            setattr(unmatched_rows_gr, colname, np.repeat(np.nan, nrow))
#
#    # remove unused columns
#    cols_to_drop = set(unmatched_rows_gr.columns).intersection(
#        {'Start_b', 'End_b', 'Distance'}
#    )
#    unmatched_rows_gr = unmatched_rows_gr[
#        unmatched_rows_gr.columns.drop(cols_to_drop).to_list()
#    ]
#    return unmatched_rows_gr
#
#
#def join_sanity_check(left_gr, right_gr):
#    assert isinstance(left_gr, pr.PyRanges)
#    assert isinstance(right_gr, pr.PyRanges)
#
#    common_cols = set.intersection(set(left_gr.columns), set(right_gr.columns))
#    assert not common_cols.difference({'Chromosome', 'Start', 'End'})
#
#    assert len(set(right_gr.columns).difference(['Chromosome', 'Start', 'End'])) > 0, (
#        f'"right_gr" Must have columns other than "Chromosome", "Start", "End".'
#    )
#
#
#def join_new(left_gr, right_gr, how='inner', find_nearest=False, merge='mean', as_gr=True):
#    assert merge in {'mean'}
#    assert how in {'inner', 'left'}
#    join_sanity_check(left_gr, right_gr)
#
#    # join
#    with warnings.catch_warnings(): 
#        warnings.simplefilter('ignore', category=FutureWarning)
#        joined_gr = left_gr.join(right_gr)
#
#    joined_df = joined_gr.df
#    added_cols = [x for x in right_gr.columns.to_list() if x not in ('Chromosome', 'Start', 'End')]
#    #if joined_gr.empty:
#        #added_cols = pd.Series(joined_gr.columns).drop(left_gr.columns).drop(['Start_b', 'End_b']).to_list()
#    #else:
#        #added_cols = joined_gr.columns.drop(left_gr.columns).drop(['Start_b', 'End_b']).to_list()
#
#    # unmatched rows
#    if how == 'left':
#        unmatched_rows_gr = extract_unmatched_rows(
#            left_gr, right_gr, joined_gr, find_nearest, added_cols,
#        )
#    elif how == 'inner':
#        unmatched_rows_gr = None
#
#    # merge
#    if joined_gr.empty:
#        matched_rows_gr = joined_gr
#    else:
#        coord_cols = np.concatenate(
#            [
#                joined_df['Chromosome'].cat.codes.values[:, np.newaxis],
#                joined_df.loc[:, ['Start', 'End']],
#            ],
#            axis=1,
#        )
#        values, counts = common.array_grouper(coord_cols, omit_values=True)
#        indexer = list(
#            itertools.chain.from_iterable(
#                itertools.repeat(val, count) for val, count in enumerate(counts)
#            )
#        )
#        grouper = joined_df.groupby(indexer)
#        with warnings.catch_warnings(): 
#            warnings.simplefilter('ignore', category=FutureWarning)
#            aggresult = grouper[added_cols].mean().reset_index(drop=True)
#
#        matched_rows = pd.concat(
#            [
#                joined_df.loc[:, left_gr.columns.to_list()].drop_duplicates(
#                    ['Chromosome', 'Start', 'End'], keep='first',
#                ).reset_index(drop=True), 
#                aggresult,
#            ], 
#            axis=1,
#        )
#        matched_rows_gr = pr.PyRanges(matched_rows)
#
#    # merge matched and nonmatched grs
#    if unmatched_rows_gr is None:
#        result = matched_rows_gr
#    else:
#        if matched_rows_gr.empty:
#            result = unmatched_rows_gr
#        else:
#            result = pr.concat([unmatched_rows_gr, matched_rows_gr]).sort()
#
#    # return
#    if not as_gr:
#        result = result.df
#
#    return result
#
#
#def join(left_gr, right_gr, how='inner', merge=None, find_nearest=False, as_gr=True, verbose=False):
#    """Args:
#        find_nearest: Match a nearest row from "right_gr" for each
#            non-overlapping row of "left_gr". Only relevant when "how" is "left".
#    Returns:
#        A pyranges.PyRanges object 
#    """
#    def merge_helper(joined_gr, added_columns, merge):
#        # modify joined_gr
#        joined_gr = joined_gr.sort()
#        joined_gr = add_new_coordinates(joined_gr, 'intersection')
#            # now 'new_Start' and 'new_End' columns are added
#        joined_gr.isec_length = joined_gr.new_End - joined_gr.new_Start
#
#        # prepare parameters
#        added_value_columns = list(set(added_columns).difference({'Start_b', 'End_b'}))
#        new_columns = ['Chromosome', 'Start', 'End', 'isec_length'] + added_value_columns
#        joined_gr_subset = joined_gr[new_columns]
#
#        # make merged annotation data
#        data = _merge_helper_make_data_vectorize(added_value_columns, joined_gr_subset, merge)
#        #data = _merge_helper_make_data_iterrows(added_value_columns, joined_gr_subset, merge)
#
#        # prepare final result
#        result_df = joined_gr.df.drop_duplicates(
#            subset=['Chromosome', 'Start', 'End'], keep='first', ignore_index=True,
#        )
#        with warnings.catch_warnings(): 
#            warnings.simplefilter('ignore', category=FutureWarning)
#            result_df.loc[:, added_value_columns] = data
#        result = pr.PyRanges(result_df, int64=False)
#        return result
#
#    def _merge_helper_make_data_iterrows(added_value_columns, joined_gr_subset, merge):
#        def _merge_values_weighted_mean(value_rowlist, lengths):
#            return np.average(np.array(value_rowlist), axis=0, weights=lengths)
#
#        def _merge_values_mean(value_rowlist, lengths):
#            return np.average(np.array(value_rowlist), axis=0, weights=None)
#
#        def _merge_values_longest_nonmerge(value_rowlist, lengths):
#            return value_rowlist[np.argmax(lengths)]
#
#        def _merge_values_longest(value_rowlist, lengths):
#            len_val_dict = dict()
#            for vals, l in zip(value_rowlist, lengths):
#                vals = tuple(vals)
#                len_val_dict.setdefault(vals, 0)
#                len_val_dict[vals] += l
#            return max(len_val_dict.items(), key=operator.itemgetter(1))[0]
#
#        if merge == 'mean':
#            values_merger = _merge_values_mean
#        elif merge == 'weighted_mean':
#            values_merger = _merge_values_weighted_mean
#        elif merge == 'longest_nonmerge':
#            values_merger = _merge_values_longest_nonmerge
#        elif merge == 'longest':
#            values_merger = _merge_values_longest
#
#        data = list()
#        for key, subiter in itertools.groupby(
#            (x[1] for x in joined_gr_subset.df.iterrows()),
#            key=(lambda row: tuple(row.iloc[:3])),
#        ):
#            value_rowlist = list()
#            lengths = list()
#            for row in subiter:
#                value_rowlist.append(row.iloc[4:])
#                lengths.append(row.iloc[3])
#
#            if len(value_rowlist) == 1:
#                data.append(value_rowlist[0])
#            else:
#                data.append(values_merger(value_rowlist, lengths))
#
#        return np.array(data)
#
#    # too slow
#    def _merge_helper_make_data_pdgroupby(added_value_columns, joined_gr_subset, merge):
#        data = list()
#        for key, subdf in joined_gr_subset.df.groupby(['Chromosome', 'Start', 'End']):
#            if subdf.shape[0] == 1:
#                data.append(subdf.iloc[0, 4:])
#            else:
#                if merge == 'mean':
#                    data.append(subdf.iloc[:, 4:].mean())
#                elif merge == 'weighted_mean':
#                    data.append(np.average(subdf.iloc[:, 4:], axis=0, weights=subdf.iloc[:, 3]))
#                elif merge == 'longest_nonmerge':
#                    data.append(subdf.iloc[np.argmax(subdf.iloc[:, 3]), 4:])
#                elif merge == 'longest':
#                    len_val_dict = {
#                        tuple(subkey): subsubdf.iloc[:, 3].sum()
#                        for subkey, subsubdf in subdf.groupby(added_value_columns)
#                    }
#                    data.append(
#                        max(len_val_dict.items(), key=operator.itemgetter(1))[0]
#                    )
#
#        return np.array(data)
#
#    # this is the fastest!
#    def _merge_helper_make_data_vectorize(added_value_columns, joined_gr_subset, merge):
#        def _merge_annotvals_weighted_mean(annotvals, lengths):
#            return np.average(np.array(annotvals), axis=0, weights=lengths)
#
#        def _merge_annotvals_mean(annotvals, lengths):
#            #return np.average(np.array(annotvals), axis=0, weights=None)
#            return np.nanmean(np.array(annotvals), axis=0)
#
#        def _merge_annotvals_longest_nonmerge(annotvals, lengths):
#            return annotvals[np.argmax(lengths)]
#
#        def _merge_annotvals_longest(annotvals, lengths):
#            len_val_dict = dict()
#            for vals, l in zip(annotvals, lengths):
#                len_val_dict.setdefault(vals, 0)
#                len_val_dict[vals] += l
#            return max(len_val_dict.items(), key=operator.itemgetter(1))[0]
#
#        coord_list = joined_gr_subset.df.apply(lambda row: tuple(row.iloc[:3]), axis=1)
#        length_list = joined_gr_subset.df.iloc[:, 3]
#        annotval_list = joined_gr_subset.df.apply((lambda row: tuple(row.iloc[4:])), axis=1)
#
#        if merge == 'mean':
#            annotvals_merger = _merge_annotvals_mean
#        elif merge == 'weighted_mean':
#            annotvals_merger = _merge_annotvals_weighted_mean
#        elif merge == 'longest_nonmerge':
#            annotvals_merger = _merge_annotvals_longest_nonmerge
#        elif merge == 'longest':
#            annotvals_merger = _merge_annotvals_longest
#
#        data = list()
#        for key, subiter in itertools.groupby(
#            zip(coord_list, length_list, annotval_list), key=operator.itemgetter(0),
#        ):
#            coords = list()
#            lengths = list()
#            annotvals = list()
#            for x in subiter:
#                coords.append(x[0])
#                lengths.append(x[1])
#                annotvals.append(x[2])
#
#            if len(coords) == 1:
#                data.extend(annotvals)
#            else:
#                data.append(annotvals_merger(annotvals, lengths))
#
#        return np.array(data)
#
#    # main
#    assert merge in ('mean', 'weighted_mean', 'first', 'longest', 'longest_nonmerge', None)
#    assert how in ('left', 'inner')
#    join_sanity_check(left_gr, right_gr)
#
#    # do join
#    #common.funclogger(1)
#    with warnings.catch_warnings(): 
#        warnings.simplefilter('ignore', category=FutureWarning)
#        joined_gr = left_gr.join(right_gr)  
#            # this is inner join
#            # setting "how='left'" does not work and results in inner join (230209)
#    #common.funclogger(2)
#    final_added_columns = [x for x in right_gr.columns.to_list() if x not in ('Chromosome', 'Start', 'End')]
#    added_columns = ['Start_b', 'End_b'] + final_added_columns
#    #if joined_gr.empty:
#        #added_columns = pd.Series(joined_gr.columns).drop(left_gr.columns).to_list()
#    #else:
#        #added_columns = joined_gr.columns.drop(left_gr.columns).to_list()
#        # This includes "Start_b" and "End_b"
#    #common.funclogger(3)
#
#    # handle unmatched rows
#    if how == 'left':
#        unmatched_rows_gr = extract_unmatched_rows(
#            left_gr, right_gr, joined_gr, find_nearest, final_added_columns,
#        )
#    elif how == 'inner':
#        unmatched_rows_gr = None
#    #common.funclogger(4)
#
#    if joined_gr.empty:
#        matched_rows_gr = joined_gr
#    else:
#        # handle matched rows - merge rows with identical (chrom, start, end)
#        if merge is None:
#            matched_rows_gr = joined_gr
#        elif merge == 'first':
#            matched_rows_gr = pr.PyRanges(
#                joined_gr.df.drop_duplicates(
#                    subset=['Chromosome', 'Start', 'End'], keep='first', ignore_index=True,
#                )
#            )
#        elif merge in ('mean', 'weighted_mean', 'longest', 'longest_nonmerge'):
#            matched_rows_gr = merge_helper(joined_gr, added_columns, merge)
#        #common.funclogger(5)
#
#        # handle matched rows - remove unused columns
#        cols_to_drop = set(matched_rows_gr.columns).intersection(
#            {'Start_b', 'End_b', 'new_Start', 'new_End', 'isec_length'}
#        )
#        matched_rows_gr = matched_rows_gr[
#            matched_rows_gr.columns.drop(cols_to_drop).to_list()
#        ]
#
#    # concat matched and unmatched rows
#    if unmatched_rows_gr is None:
#        result = matched_rows_gr
#    else:
#        if matched_rows_gr.empty:
#            result = unmatched_rows_gr
#        else:
#            result = pr.concat([unmatched_rows_gr, matched_rows_gr]).sort()
#
#    # return
#    if as_gr:
#        return result
#    else:
#        return result.df


###############################

def join_preprocess_df(df):
    df = cnvmisc.arg_into_df(df)

    leading_cols = ['Chromosome', 'Start', 'End']
    if df.columns.to_list()[:3] != leading_cols:
        df = df.loc[
            :, 
            (leading_cols + df.columns.drop(leading_cols).to_list())
        ]
    return df.reset_index(drop=True, inplace=False)


def join_df_sanitycheck(left_df, right_df, index_col, add_std):
    common_cols = ["Chromosome", "Start", "End"]

    for df in (left_df, right_df):
        if not set(common_cols).issubset(df.columns):
            raise Exception(f'Input dataframe must include columns {common_cols}')
        if index_col in df.columns:
            raise Exception(f'Input dataframe must not have a column named "{index_col}"')

    left_annot_cols = left_df.columns.drop(common_cols)
    right_annot_cols = right_df.columns.drop(common_cols)
    assert not set(left_annot_cols).intersection(set(right_annot_cols)), f'Input DataFrames has overlapping annotation columns'

    if add_std:
        std_colnames = set(x + '_std' for x in right_annot_cols)
        assert not set(left_annot_cols).intersection(std_colnames)
        assert not set(right_annot_cols).intersection(std_colnames)


def check_intv_invtlist_overlap(intv_start0, intv_end0, intvlist_start0s, intvlist_end0s):
    return np.logical_and(
        intv_end0 > intvlist_start0s,
        intv_start0 < intvlist_end0s,
    )


def compare_coords(left_df, right_df):
    pos_compare_1 = (
        right_df.End.to_numpy() 
        > left_df.Start.to_numpy()[:, np.newaxis]
    )
    pos_compare_2 = (
        right_df.Start.to_numpy() 
        < left_df.End.to_numpy()[:, np.newaxis]
    )
    compare_result = np.logical_and(pos_compare_1, pos_compare_2)

    return compare_result


def compare_coords_new(left_df, right_df):
    """Assumes input dataframes are sorted by [Start, End]"""
    left_start = left_df.Start.to_numpy()
    left_end = left_df.End.to_numpy()
    right_start = right_df.Start.to_numpy()
    right_end = right_df.End.to_numpy()
    left_nrow = left_df.shape[0]
    right_nrow = right_df.shape[0]

    compare_result = np.tile(False, (left_nrow, right_nrow))
    if left_end[-1] <= right_start[0]:
        pass
    elif right_end[-1] <= left_start[0]:
        pass
    else:
        # get matched left_df indexes
        left_idx_start0 = np.where(left_end > right_start[0])[0][0]
        where_result = np.where(left_start >= right_end[-1])[0]
        if len(where_result) == 0:
            left_idx_end0 = left_nrow
        else:
            left_idx_end0 = where_result[0]

        right_idx = 0
        for left_idx in range(left_idx_start0, left_idx_end0):
            compare1 = (left_start[left_idx] < right_end[right_idx:])
            compare2 = (left_end[left_idx] > right_start[right_idx:])
            compare = np.logical_and(compare1, compare2)

            compare_result[left_idx, right_idx:] = compare
            where_result = np.where(compare)[0]
            if len(where_result) > 0:
                right_idx += where_result[0]

    return compare_result


def fetch_nearest(joined_df, right_df, left_rowidxs, annot_columns):
    unmatched_row_selector = ~joined_df.index.isin(left_rowidxs)
    unmatched_row_coords = joined_df.loc[unmatched_row_selector, :]

    distances = calc_distances(
        unmatched_row_coords['Start'].to_numpy(), 
        unmatched_row_coords['End'].to_numpy(), 
        right_df['Start'].to_numpy(), 
        right_df['End'].to_numpy(),
    )
    nearest_right_df_indexes = np.argmin(distances, axis=1)
    joined_df.loc[unmatched_row_selector, annot_columns] = (
        right_df.iloc[nearest_right_df_indexes, :].loc[:, annot_columns].to_numpy()
    )
    return joined_df


#def join_singlechrom_dfs_new(
#    left_df, right_df, how, keep_right_coords, find_nearest, annot_columns, merge, logger,
#):
#    logger.debug(f'Sorting dataframes')
#    left_df.sort_values(['Start', 'End'], inplace=True)
#    right_df.sort_values(['Start', 'End'], inplace=True)
#
#    logger.debug(f'Comparing coordinates')
#    compare_result = compare_coords(left_df, right_df)
#    left_rowidxs, right_rowidxs = np.where(compare_result)
#
#    logger.debug(f'Subsetting right df')
#    if keep_right_coords:
#        left_subdf = left_df.iloc[left_rowidxs, :]
#        right_subdf = right_df.iloc[right_rowidxs, 3:]
#        right_subdf['overlap_length'] = calc_overlap_lengths(
#            left_subdf['Start'].to_numpy(), 
#            left_subdf['End'].to_numpy(), 
#            right_subdf['Start'].to_numpy(), 
#            right_subdf['End'].to_numpy(),
#        )
#    else:
#        right_subdf = right_df.iloc[right_rowidxs, 3:]
#    right_subdf.index = left_rowidxs
#    
#    # do merge
#    logger.debug(f'Doing merge')
#    merged_right_subdf = merge_right_subdf(right_subdf, merge, annot_columns)
#
#    # join with left_df
#    logger.debug(f'Joining with left df')
#    joined_df = left_df.join(merged_right_subdf, how=how)
#
#    # fetch nearest
#    if (how == 'left') and find_nearest:
#        logger.debug(f'Doing fetch nearest')
#        joined_df = fetch_nearest(joined_df, right_df, left_rowidxs, annot_columns)
#
#    return joined_df


def join_singlechrom_dfs(left_df, right_df, how, keep_right_coords, find_nearest, annot_columns, merge, logger, add_std, ddof):
    current_chrom = left_df.Chromosome[0]
    logger.debug(f'Beginning chromosome {current_chrom}')

#    logger.debug(f'Sorting dataframes')
#    left_df.sort_values(['Start', 'End'], inplace=True)
#    right_df.sort_values(['Start', 'End'], inplace=True)

    #logger.debug(f'Comparing coordinates')
    compare_result = compare_coords(left_df, right_df)
    left_rowidxs, right_rowidxs = np.where(compare_result)

    #logger.debug(f'Subsetting right df')
    right_subdf = right_df.iloc[right_rowidxs, 3:]
    if keep_right_coords:
        right_subdf['overlap_length'] = calc_overlap_lengths(
            left_df['Start'].iloc[left_rowidxs].to_numpy(), 
            left_df['End'].iloc[left_rowidxs].to_numpy(), 
            right_df['Start'].iloc[right_rowidxs].to_numpy(), 
            right_df['End'].iloc[right_rowidxs].to_numpy(),
        )

    right_subdf.index = left_rowidxs
    
    # do merge
    #logger.debug(f'Doing merge')
    right_subdf_merged = merge_right_subdf(right_subdf, merge, annot_columns, add_std, ddof)

    # join with left_df
    #logger.debug(f'Joining with left df')
    joined_df = left_df.join(right_subdf_merged, how=how)

    # fetch nearest
    if (how == 'left') and find_nearest:
        #logger.debug(f'Doing fetch nearest')
        joined_df = fetch_nearest(joined_df, right_df, left_rowidxs, annot_columns)

    logger.debug(f'Finished chromosome {current_chrom}')

    return joined_df


def dedup_by_index(df, index_col, keep):
    indexes = df.index.to_numpy()
    diff_indexes = np.nonzero(np.diff(indexes))[0]
    if keep == 'last':
        selector = np.concatenate([diff_indexes, [df.shape[0] - 1]])
    elif keep == 'first':
        selector = np.concatenate([[-1], diff_indexes]) + 1

    result = df.iloc[selector, :]
    result.reset_index(inplace=True, drop=True)
    return result


def merge_right_subdf_old(right_subdf, merge, annot_columns):
    if merge is None:
        return right_subdf
    elif merge == 'first':
        return dedup_by_index(right_subdf, index_col=None, keep='first')
    elif merge == 'last':
        return dedup_by_index(right_subdf, index_col=None, keep='last')
    else:
        indexes = right_subdf.index.to_numpy()

        if merge in ('mean',):
            right_subdf_merged = right_subdf.groupby(by=indexes, axis=0, sort=False).mean()
        elif merge in ('weighted_mean', 'longest'):
            if merge == 'weighted_mean':
                def subdf_handler(subdf):
                    weights = subdf['overlap_length'].to_numpy()
                    if np.isnan(weights.sum()):
                        return subdf.iloc[0, :]
                    else:
                        return subdf.apply(
                            lambda x: tools.nanaverage(x.to_numpy(), weights=weights)
                        )
            elif merge == 'longest':
                def subdf_handler(subdf):
                    idx = np.argmax(subdf['overlap_length'])
                    return subdf.iloc[idx, :]

            data_gen = (
                subdf_handler(subdf)
                for key, subdf in right_subdf.groupby(by=indexes, axis=0, sort=False)
            )
            right_subdf_merged = pd.DataFrame.from_records(data_gen)

        return right_subdf_merged


def merge_right_subdf(right_subdf, merge, annot_columns, add_std, ddof):
    if merge is None:
        return right_subdf
    elif merge == 'first':
        return right_subdf.groupby(right_subdf.index.to_numpy()).first()
    elif merge == 'last':
        return right_subdf.groupby(right_subdf.index.to_numpy()).last()
    else:
        #left_rowidxs = right_subdf.index.to_numpy()

        if merge in ('mean',):
            groupby = right_subdf.groupby(by=right_subdf.index.to_numpy(), axis=0, sort=False)
            means = groupby.mean()
            if add_std:
                stds = groupby.std(ddof=ddof).rename(columns=(lambda x: x + '_std'))
                right_subdf_merged = pd.concat([means, stds], axis=1)
            else:
                right_subdf_merged = means

        elif merge in ('weighted_mean',):
            initial_columns = right_subdf.columns.to_list()
            colname_map = {x: x + '*length' for x in initial_columns}
            for colname in initial_columns:
                right_subdf[colname_map[colname]] = right_subdf[colname] * right_subdf['overlap_length']

            groupby = right_subdf.groupby(by=right_subdf.index.to_numpy(), axis=0, sort=False)
            merged = groupby[['overlap_length'] + list(colname_map.values())].sum()

            df_data = dict()
            for key, val in colname_map.items():
                df_data[key] = merged[val] / merged['overlap_length']
            right_subdf_merged = pd.DataFrame(df_data, index=merged.index)

        elif merge in ('longest', 'shortest'):
            # keep away current index
            assert 'left_rowidxs' not in right_subdf.columns
            right_subdf.reset_index(names='left_rowidxs', inplace=True)
            groupby = right_subdf.groupby(
                by=right_subdf['left_rowidxs'].to_numpy(), axis=0, sort=False,
            )

            # do reduction
            if merge == 'longest':
                selected_row_indexes = groupby['overlap_length'].idxmax(axis=0)
            elif merge == 'shortest':
                selected_row_indexes = groupby['overlap_length'].idxmin(axis=0)
            right_subdf_merged = right_subdf.iloc[selected_row_indexes, :]

            # revert index
            right_subdf_merged.set_index('left_rowidxs')
            right_subdf_merged.index.name = None

        return right_subdf_merged


def calc_distances(starts_left, ends_left, starts_right, ends_right):
    result = np.empty((len(starts_left), len(starts_right)), dtype=int)

    left_on_left = ends_left[:, np.newaxis] <= starts_right
    right_on_left = starts_left[:, np.newaxis] >= ends_right
    overlaps = np.logical_and(
        np.logical_not(left_on_left), np.logical_not(right_on_left)
    )

    left_indexes, right_indexes = np.where(left_on_left)
    result[left_on_left] = starts_right[right_indexes] - ends_left[left_indexes]

    left_indexes, right_indexes = np.where(right_on_left)
    result[right_on_left] = starts_left[left_indexes] - ends_right[right_indexes]

    result[overlaps] = 0

    return result


def calc_overlap_lengths(starts_left, ends_left, starts_right, ends_right):
    assert len(starts_left) == len(starts_right)
    result = (
        np.min(np.array([ends_left, ends_right]), axis=0)
        - np.max(np.array([starts_left, starts_right]), axis=0)
    )
    result[result < 0] = 0
    return result


def group_df_bychrom(df):
    result = dict()
    for key, subdf in df.groupby(df['Chromosome'].to_numpy()):
        result[key] = subdf.reset_index(drop=True)
    return result


def concat_dfs_bychrom(how, sort, common_chroms, leftonly_chroms, chromdict, joined_bychrom, left_bychrom):
    if how == 'inner':
        if sort:
            sorted_chroms = sorted(
                common_chroms, 
                key=(lambda x: chromdict.contigs.index(x)),
            )
            concat_dfs = (
                joined_bychrom[x] for x in sorted_chroms
            )
        else:
            concat_dfs = iter(joined_bychrom.values())
    elif how == 'left':
        if sort:
            sorted_chroms = sorted(
                itertools.chain(common_chroms, leftonly_chroms),
                key=(lambda x: chromdict.contigs.index(x)),
            )
            concat_dfs = (
                (
                    left_bychrom[x]  
                    if x in leftonly_chroms else
                    joined_bychrom[x]
                )
                for x in sorted_chroms
            )
        else:
            concat_dfs = itertools.chain(
                iter(joined_bychrom.values()),
                (left_bychrom[chrom] for chrom in leftonly_chroms),
            )

    result = pd.concat(concat_dfs, axis=0)
    return result


def join_main_newest(
    left_df, right_df, how, merge, find_nearest, index_col, sort, chromdict, logger,
    nproc, add_std, ddof,
):
    ###
    annot_columns = right_df.columns[3:].to_list()

    ###
    logger.debug(f'Grouping dataframes by chromosome')
    left_bychrom = group_df_bychrom(left_df)
    right_bychrom = group_df_bychrom(right_df)

    left_chroms = set(left_bychrom.keys())
    right_chroms = set(right_bychrom.keys())
    common_chroms = left_chroms.intersection(right_chroms)
    leftonly_chroms = left_chroms.difference(right_chroms)
    #rightonly_chroms = right_chroms.difference(left_chroms)

    ###
    logger.debug(f'Doing join by chromosome')
    keep_right_coords = (merge in ('weighted_mean', 'longest'))
    with multiprocessing.Pool(nproc) as pool:
        args = (
            (
                left_bychrom[chrom], 
                right_bychrom[chrom], 
                how, 
                keep_right_coords,
                find_nearest,
                annot_columns,
                merge,
                logger,
                add_std,
                ddof,
            )
            for chrom in common_chroms
        )
        joined_df_list = pool.starmap(join_singlechrom_dfs, args)
    joined_bychrom = dict(zip(common_chroms, joined_df_list))

    ###
    logger.debug(f'Concatenating dataframes of all chromosomes')
    result = concat_dfs_bychrom(
        how, sort, common_chroms, leftonly_chroms, chromdict, joined_bychrom, left_bychrom,
    )
    result.reset_index(inplace=True, drop=True)

    return result


def join_newest(
    left_df, right_df, how='inner', find_nearest=False, merge=None, as_gr=False,
    add_std=False, ddof=0,
    nproc=1,
    sort=False, chromdict=None, refver=None,
    verbose=False,
):
    index_col = '__index'

    # sanity check
    assert merge in (
        'mean', 'weighted_mean', 'longest', #'longest_nonmerge', 
        'first', 'last',
        None,
    )
    assert how in ('left', 'inner')
    if sort and ((chromdict is None) and (refver is None)):
        raise Exception(f'If "sort" is True, "chromdict" or "refver" must be given')
    join_df_sanitycheck(left_df, right_df, index_col, add_std)

    # set logger
    #logger = (LOGGER_DEBUG if verbose else LOGGER_INFO)
    logger = logutils.get_funclogger(verbose)

    # arg handling
    left_df = join_preprocess_df(left_df)
    right_df = join_preprocess_df(right_df)
    if (chromdict is None) and (refver is not None):
        chromdict = refgenome.get_default_chromdict(refver)

    # join
    result = join_main_newest(left_df, right_df, how, merge, find_nearest, index_col, sort, chromdict, logger, nproc, add_std, ddof)

    # result
    if as_gr:
        result = pr.PyRanges(result)

    return result


join = join_newest





import sys
import os
import pathlib
import importlib
import importlib.resources
import importlib.metadata

import yaml

import handygenome


__version__ = importlib.metadata.version(handygenome.__name__)


# set numpy threads (https://stackoverflow.com/questions/30791550/limit-number-of-threads-in-numpy)
os.environ["OMP_NUM_THREADS"] = "1"  # export OMP_NUM_THREADS=4
os.environ["OPENBLAS_NUM_THREADS"] = "1"  # export OPENBLAS_NUM_THREADS=4 
os.environ["MKL_NUM_THREADS"] = "1"  # export MKL_NUM_THREADS=6
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"  # export VECLIB_MAXIMUM_THREADS=4
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # export NUMEXPR_NUM_THREADS=6


# USERDIR (~/.handygenome)
USERDIR = pathlib.Path.home() / '.handygenome'

CONFIGPATH = USERDIR / 'config.yaml'
with CONFIGPATH.open() as f:
    OPTION = yaml.safe_load(f)
if OPTION['python'] is None:
    OPTION['python'] = sys.executable

DATADIR = USERDIR / 'data'


#PROJECT_PATH = os.path.dirname(os.path.dirname(handygenome.__file__))  
#    # /home/users/pjh/scripts/python_genome_package_dev

# data files within top package
PKGPATH = importlib.resources.files(handygenome)
DIRS = {
    #'data': os.path.join(PROJECT_PATH, 'data'),
    #'externals': os.path.join(PROJECT_PATH, 'externals'),
    'R': PKGPATH.joinpath('Rscripts'),
    'scripts': PKGPATH.joinpath('scripts'),
    #'package': os.path.join(PROJECT_PATH, 'handygenome'),
}

#DIRS = {
#    'data': os.path.join(PROJECT_PATH, 'data'),
#    'externals': os.path.join(PROJECT_PATH, 'externals'),
#    'R': os.path.join(PROJECT_PATH, 'R'),
#    'scripts': os.path.join(PROJECT_PATH, 'scripts'),
#    'package': os.path.join(PROJECT_PATH, 'handygenome'),
#}





import os
import itertools
import functools
import pprint
import re
import collections

import pysam
import pandas as pd
import pyranges as pr

import handygenome
import handygenome.network as network
import handygenome.tools as tools
import handygenome.logutils as logutils
import handygenome.deco as deco
import handygenome.publicdb.ncbi as libncbi
import handygenome.fastahandler as fastahandler


# constants

MT_CHROMS = ('chrM', 'MT')
CHROM_PATSTRING_CHR = '(?P<chr>chr)'
CHROM_PATSTRING_NUMBER = '(?P<number>0*(?P<number_proper>[1-9][0-9]*))'
CHROM_PATSTRING_XY = '(?P<xy>[XY])'
PAT_NUMERIC_CHROM = re.compile(f'{CHROM_PATSTRING_CHR}?{CHROM_PATSTRING_NUMBER}')
PAT_ASSEMBLED_CHROM = re.compile(f'{CHROM_PATSTRING_CHR}?({CHROM_PATSTRING_NUMBER}|{CHROM_PATSTRING_XY})')
CHR1_NAME_PAT = re.compile(r'(chr)?0*1')

FASTA_URLS = {
    'GRCh37_1000genomes': 'http://ftp.1000genomes.ebi.ac.uk/vol1/ftp/technical/reference/human_g1k_v37.fasta.gz',
    'GRCh37_ucsc': 'https://hgdownload.soe.ucsc.edu/goldenPath/hg19/bigZips/analysisSet/hg19.p13.plusMT.no_alt_analysis_set.fa.gz',
    'GRCh38_1000genomes': 'http://ftp.1000genomes.ebi.ac.uk/vol1/ftp/technical/reference/GRCh38_reference_genome/GRCh38_full_analysis_set_plus_decoy_hla.fa',
}


class UnknownRefverError(Exception):
    pass


class NoChr1NameError(Exception):
    pass


# Initialize REFVERINFO from the config file

class RefverInfo:
    mandatory_keys = (
        'standard',
        'aliases',
        'parent_standard',
        'fasta_path',
        'species',
    )
    optional_keys = (
        #'assembly_report_path',
        'repeatmasker_path',
        'edited_repeat_path',

        'raw_geneset_path',
        'edited_geneset_path',

        'raw_regulatory_path',
        'edited_regulatory_path',

        'dbsnp_path',
    )
    def __init__(self, initial_data):
        self.data = list()
        for refveritem in initial_data:
            self.add_refveritem(refveritem, reset_caches=False)
        self.reset_caches()

    def reset_caches(self):
        self.set_maps()
        self.set_caches()

    def set_maps(self):
        # aliases
        self.standard_aliases_map = dict((x['standard'], x['aliases']) for x in self.data)
        # species-to-standards
        self.species_standards_map = dict()
        for x in self.data:
            self.species_standards_map.setdefault(x['species'], list())
            self.species_standards_map[x['species']].append(x['standard'])
        # default fasta
        self.standard_fastapath_map = dict(
            (x['standard'], x['fasta_path']) for x in self.data
        )
        # repeat
        self.standard_repeatmasker_map = dict(
            (x['standard'], x['repeatmasker_path']) for x in self.data
        )
        self.standard_edited_repeat_map = dict(
            (x['standard'], x['edited_repeat_path']) for x in self.data
        )
        # geneset
        self.standard_raw_geneset_map = dict(
            (x['standard'], x['raw_geneset_path']) for x in self.data
        )
        self.standard_edited_geneset_map = dict(
            (x['standard'], x['edited_geneset_path']) for x in self.data
        )
        # regulatory
        self.standard_raw_regulatory_map = dict(
            (x['standard'], x['raw_regulatory_path']) for x in self.data
        )
        self.standard_edited_regulatory_map = dict(
            (x['standard'], x['edited_regulatory_path']) for x in self.data
        )

    def set_caches(self):
        self.default_fastas = dict()
        self.default_chromdicts = dict()

        self.assemblyspecs = dict()
        self.refverspecs = dict()

        self.repeat_handles = dict()
        self.geneset_handles = dict()
        self.regulatory_handles = dict()

    def refveritem_sanitycheck(self, refveritem):
        #1
        assert isinstance(refveritem, dict), (
            f'Added reference info item must be a dict object.'
        )
        #2
        assert set(self.__class__.mandatory_keys).issubset(refveritem.keys()), (
            f'Added reference info must include these keys: '
            f'{self.__class__.mandatory_keys}'
        )
        #3
        assert refveritem['standard'] not in self.list_standards(), (
            f'"standard" of input refveritem already exists in the database.'
        )
        #4
        if refveritem['parent_standard'] is not None:
            if refveritem['species'] is None:
                raise Exception(f'When "parent_standard" is set, "species" must be also set.')
            if refveritem['parent_standard'] not in libncbi.REFSEQPATHS.get_available_refvers(refveritem['species']):
                raise Exception(f'"parent_standard" and "species" does not match according to RefSeq database.')
            if refveritem['fasta_path'] is not None:
                refverspec_fasta = RefverSpec.from_fasta_path(refveritem['fasta_path'])
                assemblyspec = AssemblySpec.from_refver(refveritem['parent_standard'], refveritem['species'])
                refverspec_asmblspec = RefverSpec.from_assemblyspec(assemblyspec)
                if not refverspec_fasta.check_length_compatibility(refverspec_asmblspec):
                    raise Exception(
                        f'For the newly added refver item, the fasta file and RefSeq assembly report do not '
                        f'coincide with the length of assembled chromosomes.\n'
                        f'refveritem: {refveritem}'
                    )
        #5
        if (refveritem['fasta_path'] is None) and (refveritem['parent_standard'] is None):
            raise Exception(f'When "parent_standard" is not set, "fasta_path" must be set.')


    # chr1 length related methods
#    @staticmethod
#    def find_chr1_name(chrom_names):
#        chr1_candidates = [x for x in chrom_names if CHR1_NAME_PAT.fullmatch(x)]
#        if len(chr1_candidates) == 1:
#            return chr1_candidates[0]
#        else:
#            raise NoChr1NameError(
#                f'Unique chromosome name which looks like chr1 could not be found.'
#            )
#
#    def find_chr1_length(self, fasta):
#        chr1_name = self.find_chr1_name(fasta.references)
#        chr1_len = fasta.lengths[fasta.references.index(chr1_name)]
#        return chr1_len

    # modifiers
    def change_fasta_path(self, refver, fasta_path):
        refveritem = self.access_refveritem(refver)
        refveritem['fasta_path'] = fasta_path
        self.reset_caches()

    def add_refveritem(self, refveritem, reset_caches=True):
        refveritem = refveritem.copy()
        for key in self.__class__.optional_keys:
            refveritem.setdefault(key, None)
        self.refveritem_sanitycheck(refveritem)
        self.data.append(refveritem)
        if reset_caches:
            self.reset_caches()

    # standardize
    def standardize(self, refver):
        for standard, alias_list in self.standard_aliases_map.items():
            if (refver == standard) or (refver in alias_list):
                return standard
        raise UnknownRefverError(f'Unknown reference version: {refver}')

    # refveritem accessor
    def access_refveritem(self, refver):
        standard = self.standardize(refver)
        for x in self.data:
            if x['standard'] == standard:
                return x
        raise Exception(f'Could not find a refveritem corresponding to the refver "{refver}"')

    # convenience methods
    def list_standards(self):
        return [x['standard'] for x in self.data]

    def list_known_refvers(self):
        result = list()
        for x in self.data:
            result.append(x['standard'])
            result.extend(x['aliases'])
        return result

    def list_known_species(self):
        return tuple(set(x['species'] for x in self.data))

    def get_parent_standard(self, refver):
        refveritem = self.access_refveritem(refver)
        return refveritem['parent_standard']

    # map fetchers
    def get_aliases(self, refver):
        standard = self.standardize(refver)
        alias_list = self.standard_aliases_map[standard]
        return [standard] + alias_list

    def find_species(self, refver):
        standard = self.standardize(refver)
        for key, val in self.species_standards_map.items():
            if standard in val:
                return key
        raise Exception(
            f'A standard refver "{standard}" could not be found from a search over "species_standards_map"'
        )

    # datafile path fetcher
    def get_datafile_path(self, refver, map_name, refseq_pathfunc_name):
        standard = self.standardize(refver)
        config_path = getattr(self, map_name)[standard]
        if config_path is None:
            refveritem = self.access_refveritem(standard)
            assert refveritem['parent_standard'] is not None
            species = refveritem['species']
            result = getattr(libncbi, refseq_pathfunc_name)(standard, species)
        else:
            result = config_path
        return result

    def get_fasta_path(self, refver):
        return self.get_datafile_path(
            refver, 'standard_fastapath_map', 'get_edited_fasta_path',
        )

    def get_repeatmasker_path(self, refver):
        return self.get_datafile_path(
            refver, 'standard_repeatmasker_map', 'get_repeatmasker_path',
        )

    def get_repeat_path(self, refver):
        standard = self.standardize(refver)
        config_path = self.standard_edited_repeat_map[standard]
        if config_path is None:
            return None
        else:
            return config_path

    def get_geneset_path(self, refver):
        standard = self.standardize(refver)
        config_path = self.standard_edited_geneset_map[standard]
        if config_path is None:
            return None
        else:
            return config_path

    def get_regulatory_path(self, refver):
        standard = self.standardize(refver)
        config_path = self.standard_edited_regulatory_map[standard]
        if config_path is None:
            return None
        else:
            return config_path

    # cache fetchers
    def get_default_fasta(self, refver):
        standard = self.standardize(refver)
        if standard not in self.default_fastas.keys():
            self.default_fastas[standard] = pysam.FastaFile(self.get_fasta_path(standard))
        return self.default_fastas[standard]

    def get_default_chromdict(self, refver):
        standard = self.standardize(refver)
        if standard not in self.default_chromdicts.keys():
            self.default_chromdicts[standard] = ChromDict.from_fasta(self.get_default_fasta(standard))
        return self.default_chromdicts[standard]

    def get_assemblyspec(self, refver):
        parent_standard = self.access_refveritem(refver)['parent_standard']
        if parent_standard is None:
            return None
            #raise Exception(f'AssemblySpec not available for this refver.')

        if parent_standard not in self.assemblyspecs.keys():
            species = self.find_species(parent_standard)
            self.assemblyspecs[parent_standard] = AssemblySpec.from_refver(parent_standard, species)

        return self.assemblyspecs[parent_standard]

    def get_refverspec(self, refver):
        standard = self.standardize(refver)

        if standard not in self.refverspecs.keys():
            assemblyspec = self.get_assemblyspec(standard)
            if assemblyspec is not None:
                self.refverspecs[standard] = RefverSpec.from_assemblyspec(assemblyspec)
            else:
                fasta_path = self.get_fasta_path(standard)
                if fasta_path is None:
                    self.refverspecs[standard] = None
                else:
                    self.refverspecs[standard] = RefverSpec.from_fasta_path(fasta_path)

        return self.refverspecs[standard]

    def get_geneset_handle(self, refver):
        standard = self.standardize(refver)
        if standard not in self.geneset_handles.keys():
            self.geneset_handles[standard] = pysam.TabixFile(
                self.get_geneset_path(standard),
                parser=pysam.asGTF(),
            )
        return self.geneset_handles[standard]

    def get_regulatory_handle(self, refver):
        standard = self.standardize(refver)
        if standard not in self.regulatory_handles.keys():
            self.regulatory_handles[standard] = pysam.TabixFile(
                self.get_regulatory_path(standard),
                parser=pysam.asGTF(),
            )
        return self.regulatory_handles[standard]

    def get_repeat_handle(self, refver):
        standard = self.standardize(refver)
        if standard not in self.repeat_handles.keys():
            self.repeat_handles[standard] = pysam.TabixFile(
                self.get_repeat_path(standard),
                parser=pysam.asBed(),
            )
        return self.repeat_handles[standard]



# RefverDict

class RefverDict(collections.UserDict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        standard_list = REFVERINFO.list_standards()
        if not set(self.keys()).issubset(standard_list):
            raise Exception(
                f'RefverDict construction keys must be restricted to: '
                f'{standard_list}'
            )

    def __getitem__(self, key):
        key = REFVERINFO.standardize(key)
        try:
            result = super().__getitem__(key)
        except KeyError:
            raise Exception(f'Input reference version is not available.')

        return result

    def __contains__(self, key):
        key = REFVERINFO.standardize(key)
        return super().__contains__(key)

    def get_valid_keys(self):
        result = list()
        for standard in self.keys():
            result.append(standard)
            result.extend(REFVERINFO.standard_aliases_map[standard])
        return result

    ############################
    # candidates be deprecated #
    @classmethod
    def standardize(cls, refver):
        return REFVERINFO.standardize(refver)

    @classmethod
    def get_aliases(cls, refver):
        return REFVERINFO.get_aliases(refver)
    ############################



#DEFAULT_FASTA_PATHS = RefverDict(handygenome.OPTION['default_fasta'])
#DEFAULT_FASTAS = RefverDict(
#    {
#        refver: pysam.FastaFile(path) 
#        for refver, path in DEFAULT_FASTA_PATHS.items()
#    }
#)



# chromdict

class ChromDict(collections.OrderedDict):
    ################
    # initializers #
    ################

    @classmethod
    def _init_from_pysamobj(cls, pysamobj):
        result = cls()
        for chrom, length in zip(pysamobj.references, pysamobj.lengths):
            result[chrom] = length
        return result

    @classmethod
    def from_fasta_path(cls, fasta_path):
        with pysam.FastaFile(fasta_path) as pysamobj:
            return cls._init_from_pysamobj(pysamobj)

    @classmethod
    def from_fasta(cls, fasta):
        return cls._init_from_pysamobj(fasta)

    @classmethod
    def from_bam_path(cls, bam_path):
        with pysam.AlignmentFile(bam_path) as pysamobj:
            return cls._init_from_pysamobj(pysamobj)

    @classmethod
    def from_bam(cls, bam):
        return cls._init_from_pysamobj(bam)

    @classmethod
    def from_bamheader(cls, bamheader):
        return cls._init_from_pysamobj(bamheader)

    @classmethod
    def from_vcfheader(cls, vcfheader):
        result = cls()
        for contig in vcfheader.contigs.values():
            result[contig.name] = contig.length
        return result

    @classmethod
    def from_refver(cls, refver):
        fasta_path = REFVERINFO.get_fasta_path(refver)
        with pysam.FastaFile(fasta_path) as fasta:
            return cls.from_fasta(fasta)

    @classmethod
    def from_custom(cls, custom):
        """Args:
            custom: {
                'contigs': ['contig1', 'contig2', ...], 
                'lengths': [length1, length2, ...],
            }
        """
        result = cls()
        for chrom, length in zip(custom['contigs'], custom['lengths']):
            result[chrom] = length
        return result

    ##############
    # properties #
    ##############

    @property
    def contigs(self):
        return list(self.keys())

    @property
    def lengths(self):
        return list(self.values())

#    @deco.get_deco_num_set_differently(
#        (
#            'fasta_path', 'fasta', 'bam_path', 'bam', 
#            'vcfheader', 'bamheader', 'custom', 'refver',
#        ), 
#        1,
#    )
#    def __init__(self, fasta_path=None, fasta=None, bam_path=None, bam=None, 
#                 vcfheader=None, bamheader=None, custom=None, refver=None):
#        """
#        Args:
#            fasta: pysam.FastaFile object
#            bam: pysam.AlignmentFile object
#            vcfheader: pysam.VariantHeader object
#            bamheader: pysam.AlignmentHeader object
#            custom: {'contigs': ['contig1', 'contig2', ...], 
#                     'lengths': [length1, length2, ...] }
#        """
#
#        # set self dict
#        if vcfheader is not None:
#            for contig in vcfheader.contigs.values():
#                self[contig.name] = contig.length
#        elif custom is not None:
#            for chrom, length in zip(custom['contigs'], custom['lengths']):
#                self[chrom] = length
#        else:
#            if fasta_path is not None:
#                wrapper = pysam.FastaFile(fasta_path)
#            elif fasta is not None:
#                wrapper = fasta
#            elif bam_path is not None:
#                wrapper = pysam.AlignmentFile(bam_path)
#            elif bam is not None:
#                wrapper = bam
#            elif bamheader is not None:
#                wrapper = bamheader
#            elif refver is not None:
#                wrapper = DEFAULT_FASTAS[refver]
#    
#            for chrom, length in zip(wrapper.references, wrapper.lengths):
#                self[chrom] = length
#    
#            if any((x is not None) for x in (fasta_path, bam_path)):
#                wrapper.close()
#
#        # set contigs, lengths
#        self.contigs = list(self.keys())
#        self.lengths = list(self.values())

    ##########
    # others #
    ##########

    def get_cumpos0(self, chrom, pos0):
        chromidx = self.contigs.index(chrom)
        if chromidx == 0:
            return pos0
        else:
            return pos0 + sum(self.lengths[:chromidx])

    @functools.cached_property
    def is_chr_prefixed(self):
        relevant_chroms = [
            x for x in self.contigs 
            if PAT_ASSEMBLED_CHROM.fullmatch(x) is not None
        ]
        startswith_chr = [x.startswith('chr') for x in relevant_chroms]
        if all(startswith_chr):
            return True
        elif not any(startswith_chr):
            return False
        else:
            raise Exception(f'Chromosome names are inconsistent of whether prefixed with "chr"')

    @functools.cached_property
    def XY_names(self):
        def helper(name, xy):
            assert xy in ('X', 'Y')
            if name in self.contigs:
                return name
            else:
                raise Exception(f'No {xy} chromosome name detected')
            
        if self.is_chr_prefixed:
            X = helper('chrX', 'X')
            Y = helper('chrY', 'Y')
        else:
            X = helper('X', 'X')
            Y = helper('Y', 'Y')

        return X, Y

    @functools.cached_property
    def assembled_chroms(self):
        return [x for x in self.contigs if PAT_ASSEMBLED_CHROM.fullmatch(x) is not None]

    def to_gr(self, assembled_only=True, as_gr=True):
        result = pd.DataFrame({
            'Chromosome': self.contigs,
            'Start': 0,
            'End': self.lengths,
        })
        if assembled_only:
            selector = result['Chromosome'].apply(
                lambda x: PAT_ASSEMBLED_CHROM.fullmatch(x) is not None
            )
            result = result.loc[selector, :]

        if as_gr:
            return pr.PyRanges(result)
        else:
            return result

#    def to_interval_list(self):
#        intvlist = libinterval.IntervalList()
#        for contig, length in zip(self.contigs, self.lengths):
#            interval = libinterval.Interval(contig, start0=0, end0=length)
#            intvlist.append(interval)
#
#        return intvlist

    def get_chrom_indexes(self, chroms):
        if np.isscalar(chroms):
            return self.contigs.index(chroms)
        else:
            where = np.where(
                np.array(chroms)[:, np.newaxis] == np.array(self.contigs)
            )
            assert len(set(where[0])) == len(chroms), f'Unknown chromosome names are included.'
            result = [
                tuple(subiter)[0][1] for key, subiter in itertools.groupby(
                    zip(where[0], where[1]), key=operator.itemgetter(0)
                )
            ]
            return result

    def get_chrompos_sortkey(self, chroms, start0s=None, end0s=None):
        assert not ((start0s is None) and (end0s is not None))

        chrom_indexes = self.get_chrom_indexes(chroms)
        if start0s is None:
            sortkey = np.lexsort([chrom_indexes])
        else:
            if end0s is None:
                sortkey = np.lexsort([start0s, chrom_indexes])
            else:
                sortkey = np.lexsort([end0s, start0s, chrom_indexes])

        return sortkey

    def sort_chrompos(self, chroms, start0s=None, end0s=None):
        assert not ((start0s is None) and (end0s is not None))

        # arg handling
        chroms = np.array(chroms)
        if start0s is not None:
            start0s = np.asarray(start0s)
        if end0s is not None:
            end0s = np.asarray(end0s)

        # result
        sortkey = self.get_chrompos_sortkey(chroms, start0s, end0s)
        if start0s is None:
            return chroms[sortkey]
        else:
            if end0s is None:
                return chroms[sortkey], start0s[sortkey]
            else:
                return chroms[sortkey], start0s[sortkey], end0s[sortkey]


#DEFAULT_CHROMDICTS = RefverDict({
#    key: ChromDict(fasta=val) for key, val in DEFAULT_FASTAS.items()
#    if key != 'GRCh37_hs37d5'
#})


############################
# Assemblyspec and related #
############################

def normalize_chrom(chrom, strip_chr=False):
    """Examples:
        chr1 -> chr1
        chrX -> chrX
        chr001 -> chr1
        1 -> chr1
        X -> chrX
        MT -> chrM
        NC_000001.10 -> NC_000001.10
        chr19_gl000209_random -> chr19_gl000209_random
    """
    mat = PAT_ASSEMBLED_CHROM.fullmatch(chrom)
    if mat is None:
        if chrom in MT_CHROMS:
            if strip_chr:
                return 'MT'
            else:
                return 'chrM'
        else:  # e.g. chr19_gl000209_random, NC_000001.10
            return chrom
    else:  # e.g. chr1, 1, X, chrY, chr001
        if mat.group('xy') is None:
            if strip_chr:
                return mat.group('number_proper')
            else:
                return 'chr' + mat.group('number_proper')
        else:
            assert mat.group('number') is None
            if strip_chr:
                return mat.group('xy')
            else:
                return 'chr' + mat.group('xy')


class AssemblySpec:
    """Attributes:
        data: { 
            'length': [249250621, 243199373, ...],
            'role': ['assembled-molecule', ... 'unlocalized-scaffold', ... ],
            'default': ['1', '2', ... 'HSCHR1_RANDOM_CTG5', ... ],
            'ucsc': ['chr1', 'chr2', ... 'chr1_gl000191_random', ... ],
            'nochr_plus_genbank': ['1', '2', ... 'GL000191.1', ... ],
            'genbank': ['CM000663.1', 'CM000664.1', ... ],
            'refseq': ['NC_000001.11', 'NC_000002.12', ... ],
        }
        dicts: {('ucsc', 'genbank'): dict, ... }
        chromdicts: {'ucsc': ChromDict object, ... }
        versions: All self.data keys except 'length' and 'role'
    """
    assembly_report_keydict = {
        'default': 'Sequence-Name',
        'ucsc': 'UCSC-style-name',
        'genbank': 'GenBank-Accn',
        'refseq': 'RefSeq-Accn',
        'length': 'Sequence-Length',
        'role': 'Sequence-Role',
    }
    def __init__(self, data):
        """
        Args:
            data: {
                'ucsc': ['chr1', 'chr2', ... ], 
                'refseq': ['NC_000001.11', 'NC_000002.12', ... ],
                ...
                }
        """

        self.data = data
        self.dicts = dict()
        self.chromdicts = dict()

        for key1, key2 in itertools.permutations(data.keys(), 2):
            dic = dict()
            chromdict_data = {'contigs': list(), 'lengths': list()}
            for val1, val2 in zip(self.data[key1], self.data[key2]):
                if val1 is not None:
                    dic[val1] = val2
                    if key2 == 'length':
                        chromdict_data['contigs'].append(val1)
                        chromdict_data['lengths'].append(val2)

            self.dicts[(key1, key2)] = dic
            if key2 == 'length':
                self.chromdicts[key1] = ChromDict(custom=chromdict_data)

        self.versions = list(
            set(self.data.keys()).difference(('length', 'role')))

    @classmethod
    def from_refver(cls, refseq_refver, species):
        #species = REFVERINFO.find_species(refver)
        assemblyfile_path = libncbi.get_assemblyfile_path(
            refseq_refver, species, force_download=False,
        )
        with open(assemblyfile_path, 'r') as infile:
            result = cls.from_assembly_report_file(infile)

        return result

    @classmethod
    def from_assembly_report_file(cls, stream):
        parsed_data = cls.parse_assembly_report(stream)
        result = cls(parsed_data)
        return result

    @classmethod
    def parse_assembly_report(cls, stream):
        data = {'length': list(), 'default': list(), 'ucsc': list(),
                'nochr_plus_genbank': list(), 'role': list(),
                'genbank': list(), 'refseq': list()}

        record_start = False
        for line in stream:
            if not record_start:
                if line.startswith('# Sequence-Name'):
                    keys = tools.get_linesp(re.sub('^#\s*', '', line))
                    record_start = True
                    continue
            else:
                linesp = tools.get_linesp(line)
                linedict = dict(
                    zip(
                        keys, 
                        [(None if x == 'na' else x) for x in linesp]
                    )
                )

                default = linedict[cls.assembly_report_keydict['default']]
                ucsc = linedict[cls.assembly_report_keydict['ucsc']]
                genbank = linedict[cls.assembly_report_keydict['genbank']]
                refseq = linedict[cls.assembly_report_keydict['refseq']]
                length = int(linedict[cls.assembly_report_keydict['length']])
                role = linedict[cls.assembly_report_keydict['role']]

                #assert genbank is not None

                data['default'].append(default)
                data['ucsc'].append(ucsc)
                data['genbank'].append(genbank)
                data['refseq'].append(refseq)
                data['length'].append(length)
                data['role'].append(role)

                if ucsc is None:
                    data['nochr_plus_genbank'].append(genbank)
                elif ucsc == 'chrM':
                    data['nochr_plus_genbank'].append('MT')
                else:
                    mat = PAT_ASSEMBLED_CHROM.fullmatch(ucsc)
                    if mat is None:
                        data['nochr_plus_genbank'].append(genbank)
                    else:
                        data['nochr_plus_genbank'].append(mat.group(2))

        return data

    def __repr__(self):
        tmp_keys = list()
        tmp_vals = list()
        for key, val in self.data.items():
            tmp_keys.append(key)
            tmp_vals.append(val)

        show_result = list()
        for tup in zip(*tmp_vals):
            show_result.append(str(dict(zip(tmp_keys, tup))))

        return '\n'.join(show_result)

    def convert(self, in_chrom, out_version):
        assert out_version in self.versions, (
            f'"out_version" argument must be one of {self.versions}')

        absent = True
        for version, names in self.data.items():
            if version in ('length', 'role'):
                continue
            if in_chrom in names:
                absent = False
                if version == out_version:
                    result = in_chrom
                else:
                    result = self.dicts[(version, out_version)][in_chrom]
                break

        if absent:
            raise Exception(f'Input chrom name "{in_chrom}" is not included '
                            f'in the database.')

        return result

    def get_alias_groups(self):
        data_keys = (
            'ucsc', 
            'genbank', 
            'refseq',
            'default', 
            'nochr_plus_genbank', 
        )
        alias_groups = list()
        for idx in range(len(self.data['length'])):
            group = set()
            for key in data_keys:
                chrom = self.data[key][idx]
                if chrom is None:
                    continue

                norm_chrom = normalize_chrom(chrom, strip_chr=False)
                norm_chrom_nochr = normalize_chrom(chrom, strip_chr=True)
                group.add(chrom)
                group.add(norm_chrom)
                group.add(norm_chrom_nochr)

            alias_groups.append(group)
        return alias_groups

    def get_chrom_converter(self, contigs, lengths):
        # check ChromLenghtHash
        hash_assemblyspec = RefverSpec.from_assemblyspec(self)
        hash_inputdata = RefverSpec.from_data(contigs, lengths)
        if not hash_assemblyspec.check_length_compatibility(hash_inputdata):
            raise Exception(f'Input contig names and lenghts does not match assemblyspec.')

        # main
        alias_groups = self.get_alias_groups()
        converter = dict()
        for chrom in contigs:
            for group in alias_groups:
                if chrom in group:
                    for alias in group:
                        assert alias not in converter
                        converter[alias] = chrom
                    continue
        return converter


#ASSEMBLYFILE_URLS = RefverDict(libncbi.collect_assemblyfile_urls())

#ASSEMBLYFILE_PATHS = {
#    key: os.path.join(ASSEMBLYFILE_DIR, os.path.basename(val))
#    for (key, val) in ASSEMBLYFILE_URLS.items()
#}


#def get_assemblyspec_data(refver):
#    assemblyfile_url = ASSEMBLYFILE_URLS[refver]
#    assembly_report_path = ASSEMBLYFILE_PATHS[refver]
#    if not os.path.exists(assembly_report_path):
#        network.download(assemblyfile_url, assembly_report_path)
#
#    return parse_assembly_report(assembly_report_path)
#
#
#SPECS = RefverDict(
#    {
#        refver: AssemblySpec(get_assemblyspec_data(refver))
#        for refver in ASSEMBLYFILE_URLS.keys()
#    }
#)


class RefverSpec:
    def __init__(self, chrom_list, length_list):
        self.lendict = dict((x, y) for (x, y) in zip(chrom_list, length_list))
        self.normchrom_lendict = self.make_normchrom_lendict(chrom_list, length_list)

    @staticmethod
    def make_normchrom_lendict(chrom_list, length_list):
        normchrom_lendict = dict()
        modified_chrom_groups = dict()
        for chrom, length in zip(chrom_list, length_list):
            modified_chrom = normalize_chrom(chrom, strip_chr=False)
            normchrom_lendict[modified_chrom] = length
            modified_chrom_groups.setdefault(modified_chrom, set())
            modified_chrom_groups[modified_chrom].add(
                (chrom, length)
            )

        # sanity check
        for modified_chrom, tuples in modified_chrom_groups.items():
            lengths = set(x[1] for x in tuples)
            if len(lengths) != 1:
                raise Exception(
                    f'Modified chrom names overlap but lengths are discordant:'
                    f'modified chrom: {modified_chrom}, original chrom - length pairs: {tuples}'
                )

        return normchrom_lendict

    @classmethod
    def from_data(cls, chrom_list, length_list):
        #lendict = cls.make_lendict(chrom_list, length_list)
        return cls(chrom_list, length_list)

    @classmethod
    def from_fasta(cls, fasta):
        return cls.from_data(fasta.references, fasta.lengths)

    @classmethod
    def from_fasta_path(cls, fasta_path):
        with pysam.FastaFile(fasta_path) as fasta:
            return cls.from_fasta(fasta)

    @classmethod
    def from_assemblyspec(cls, assemblyspec):
        assert isinstance(assemblyspec, AssemblySpec)

        chrom_list = list(
            itertools.chain(
                assemblyspec.data['default'],
                assemblyspec.data['ucsc'],
                assemblyspec.data['nochr_plus_genbank'],
                assemblyspec.data['genbank'],
                assemblyspec.data['refseq'],
            )
        )
        length_list = assemblyspec.data['length'] * 5
        # remove None chroms
        notnone_indexes = tuple((x is not None) for x in chrom_list)
        chrom_list = list(itertools.compress(chrom_list, notnone_indexes))
        length_list = list(itertools.compress(length_list, notnone_indexes))

        return cls.from_data(chrom_list, length_list)

    @staticmethod
    def compare_lendicts(self_lendict, other_lendict):
        self_chroms = set(self_lendict.keys())
        other_chroms = set(other_lendict.keys())

        self_only_chroms = self_chroms.difference(other_chroms)
        other_only_chroms = other_chroms.difference(self_chroms)
        common_chroms = self_chroms.intersection(other_chroms)
        length_matches = all(
            (self_lendict[x] == other_lendict[x])
            for x in common_chroms
        )
        return length_matches, common_chroms, self_only_chroms, other_only_chroms

    def check_length_compatibility(self, other):
        """Used for parent standard identification"""
        assert isinstance(other, RefverSpec)
        length_matches, common_chroms, self_only_chroms, other_only_chroms = (
            self.compare_lendicts(self.normchrom_lendict, other.normchrom_lendict)
        )
        return length_matches

    def check_issubset(self, other):
        assert isinstance(other, RefverSpec)
        length_matches, common_chroms, self_only_chroms, other_only_chroms = (
            self.compare_lendicts(self.lendict, other.lendict)
        )
        is_subset = ((len(self_only_chroms) == 0) and length_matches)
        return is_subset, other_only_chroms

    def get_hash(self, chroms=None):
        if chroms is None:
            chrom_len_tuples = iter(self.lendict.items())
        else:
            chrom_len_tuples = ((x, self.lendict[x]) for x in chroms)
        return tuple(sorted(chrom_len_tuples, key=(lambda x: x[0])))

    def compare(self, other):
        assert isinstance(other, RefverSpec)
        common_chroms = set(self.lendict.keys()).intersection(other.lendict.keys())
        return self.get_hash(common_chroms) == other.get_hash(common_chroms)
                
                    


####################
# Setup REFVERINFO #
####################

REFVERINFO = RefverInfo(handygenome.OPTION['refverinfo'])


###########################################################################
# convenience functions for utilizing and modifying the REFVERINFO object #
###########################################################################

def standardize_refver(refver):
    REFVERINFO.standardize(refver)

def add_refveritem(refveritem):
    REFVERINFO.add_refveritem(refveritem, reset_caches=True)
    

def change_fasta_path(refver, fasta_path):
    REFVERINFO.change_fasta_path(refver, fasta_path)


def list_known_refvers():
    return REFVERINFO.list_known_refvers()


def get_fasta_path(refver):
    return REFVERINFO.get_fasta_path(refver)


def get_default_fasta(refver):
    return REFVERINFO.get_default_fasta(refver)


def get_default_chromdict(refver):
    return REFVERINFO.get_default_chromdict(refver)


def get_assemblyspec(refver):
    return REFVERINFO.get_assemblyspec(refver)


def get_geneset_handle(refver):
    return REFVERINFO.get_geneset_handle(refver)
    

def get_regulatory_handle(refver):
    return REFVERINFO.get_regulatory_handle(refver)


def get_repeat_handle(refver):
    return REFVERINFO.get_repeat_handle(refver)


###############################
# reference version inference #
###############################

@deco.get_deco_num_set_differently(('chromdict', 'vcfheader', 'bamheader'), 1)
def infer_refver(chromdict=None, vcfheader=None, bamheader=None):
    if chromdict is not None:
        return infer_refver_chromdict(chromdict)
    elif vcfheader is not None:
        return infer_refver_vcfheader(vcfheader)
    elif bamheader is not None:
        return infer_refver_bamheader(bamheader)


def infer_parent_standard(contigs, lengths):
    query_refverspec = RefverSpec.from_data(contigs, lengths)


def infer_refver_base(contigs, lengths):
    query_hash = RefverSpec.from_data(contigs, lengths)
    compare_results = list()
    for standard in REFVERINFO.list_standards():
        target_hash = REFVERINFO.get_refverspec(standard)
        if target_hash is None:
            continue
        is_subset, other_only_chroms = target_hash.check_issubset(query_hash)
        compare_results.append((standard, is_subset, other_only_chroms))

    subset_refvers = [x for x in compare_results if x[1]]
    if len(subset_refvers) == 1:
        return subset_refvers[0][0]
    elif len(subset_refvers) > 1:
        minimal_descrepancy_refvers = tools.multi_min(
            subset_refvers, key=(lambda x: len(x[2]))
        )
        if len(minimal_descrepancy_refvers) == 1:
            return minimal_descrepancy_refvers[0][0]
        else:
            raise Exception(f'Cannot break tie: {subset_refvers}')
    elif len(subset_refvers) == 0:
        return None
            
        

    

    elif len(candidates) > 1:
        raise Exception(f'Matches assembled chromosome lengths with more than one reference genomes({candidates})')
    else:
        return candidates[0]

    if len(candidates) == 0:
        raise Exception(f'Cannot infer refver: unknown chromosome lengths')
    elif len(candidates) > 1:
        raise Exception(f'Matches assembled chromosome lengths with more than one reference genomes({candidates})')
    else:
        return candidates[0]


def infer_refver_chromdict(chromdict):
    return infer_refver_base(chromdict.contigs, chromdict.lengths)


def infer_refver_vcfheader(vcfheader):
    contigs = list()
    lengths = list()
    for contig in vcfheader.contigs.values():
        contigs.append(contig.name)
        lengths.append(contig.length)
    return infer_refver_base(contigs, lengths)


def infer_refver_vr(vr):
    return infer_refver_vcfheader(vr.header)


def infer_refver_fasta(fasta):
    return infer_refver_pysamwrapper(fasta)


def infer_refver_bamheader(bamheader):
    return infer_refver_pysamwrapper(bamheader)


def infer_refver_pysamwrapper(wrapper):
    return infer_refver_base(wrapper.references, wrapper.lengths)


##################################################
# handling of pseudoautosomal region & runs of N #
##################################################

N_REGIONS_DIR = os.path.join(handygenome.DIRS['data'], 'N_region')
if not os.path.exists(N_REGIONS_DIR):
    os.mkdir(N_REGIONS_DIR)

PARS = RefverDict(
    # 0-based half-open system
    # ref: https://www.ncbi.nlm.nih.gov/grc/human
    {  
        'GRCh37': {
            'PAR1_X': ('X', 60_000, 2_699_520),
            'PAR2_X': ('X', 154_931_043, 155_260_560),
            'PAR1_Y': ('Y', 10_000, 2_649_520),
            'PAR2_Y': ('Y', 59_034_049, 59_373_566),
        }, 
        'GRCh38': {
            'PAR1_X': ('X', 10_000, 2_781_479),
            'PAR2_X': ('X', 155_701_382, 156_030_895),
            'PAR1_Y': ('Y', 10_000, 2_781_479),
            'PAR2_Y': ('Y', 56_887_902, 57_217_415),
        },
    }
)

PAR_GRS = dict()
for key, val in PARS.items():
    df = pd.DataFrame.from_records(
        iter(val.values()), 
        columns=('Chromosome', 'Start', 'End'),
    )
    gr = pr.PyRanges(df).sort()
    PAR_GRS[key] = gr
PAR_GRS = RefverDict(PAR_GRS)


def get_par_gr(refver):
    chromdict = ChromDict.from_refver(refver)
    par_gr = PAR_GRS[refver].copy()  # X, Y
    if chromdict.is_chr_prefixed:
        par_gr.Chromosome = 'chr' + par_gr.Chromosome
    return par_gr


def make_N_region_gr(fasta):
    pat = re.compile('N+')
    chroms = list()
    start0s = list()
    end0s = list()
    for chrom in fasta.references:
        seq = fasta.fetch(chrom)
        offset = 0
        while True:
            mat = pat.search(seq)
            if mat is None:
                break
            else:
                span = mat.span()
                chroms.append(chrom)
                start0s.append(offset + span[0])
                end0s.append(offset + span[1])

                offset += span[1]
                seq = seq[span[1]:]

    return pr.PyRanges(chromosomes=chroms, starts=start0s, ends=end0s)


def get_N_regionfile_path(refver):
    refver = RefverDict.standardize(refver)
    return os.path.join(N_REGIONS_DIR, f'{refver}.N_regions.tsv.gz')


def write_N_regionfile(refver):
    fasta = get_default_fasta(refver)
    gr = make_N_region_gr(fasta)
    N_regionfile_path = get_N_regionfile_path(refver)
    gr.df.to_csv(N_regionfile_path, sep='\t', index=False)


def get_N_region_gr(refver):
    N_regionfile_path = get_N_regionfile_path(refver)
    if not os.path.exists(N_regionfile_path):
        print('Creating a N region file. It may take a few minutes.')
        write_N_regionfile(refver)
    return pr.PyRanges(
        pd.read_csv(N_regionfile_path, sep='\t', header=0, dtype={'Chromosome': str})
    )
        

